# Copyright Modal Labs 2023
import contextlib
import inspect
import logging
import typing
from collections import Counter, defaultdict
from collections.abc import Awaitable
from typing import Any, Callable

import grpclib.server
from grpclib import GRPCError, Status

from modal.config import logger

if typing.TYPE_CHECKING:
    from test.conftest import MockClientServicer


def patch_mock_servicer(cls):
    """Adds an `.intercept()` context manager method

    This allows for context-local tracking and assertions of all calls
    performed on the servicer during a context, e.g.:

    ```python notest
    with servicer.intercept() as ctx:
        await some_complex_method()
    assert ctx.calls == [("SomeMethod", MyMessage(foo="bar"))]
    ```
    Also allows to set a predefined queue of responses, temporarily replacing
    a mock servicer's default responses for a method:

    ```python notest
    with servicer.intercept() as ctx:
        ctx.add_response("SomeMethod", [
            MyResponse(bar="baz")
        ])
        ctx.add_response("SomeMethod", [
            MyResponse(bar="baz2")
        ])
        await service_stub.SomeMethod(Empty())  # receives MyResponse(bar="baz")
        await service_stub.SomeMethod(Empty())  # receives MyResponse(bar="baz2")
    ```

    Also patches all unimplemented abstract methods in a mock servicer with default error implementations.
    """

    async def fallback(self, stream) -> None:
        raise GRPCError(Status.UNIMPLEMENTED, "Not implemented in mock servicer " + repr(cls))

    @contextlib.contextmanager
    def intercept(servicer):
        ctx = InterceptionContext(servicer)
        servicer.interception_context = ctx
        yield ctx
        ctx._assert_responses_consumed()
        servicer.interception_context = None

    cls.intercept = intercept
    cls.interception_context = None

    def patch_grpc_method(method_name, original_method):
        async def patched_method(servicer_self, stream):
            try:
                ctx = servicer_self.interception_context
                if ctx:
                    intercepted_stream = await InterceptedStream(ctx, method_name, stream).initialize()
                    custom_responder = ctx._next_custom_responder(method_name, intercepted_stream.request_message)
                    if custom_responder:
                        return await custom_responder(servicer_self, intercepted_stream)
                    else:
                        # use default servicer, but intercept messages for assertions
                        return await original_method(servicer_self, intercepted_stream)
                else:
                    return await original_method(servicer_self, stream)
            except GRPCError:
                raise
            except Exception:
                logger.exception("Error in mock servicer responder:")
                raise

        return patched_method

    # Fill in the remaining methods on the class
    for name in dir(cls):
        method = getattr(cls, name)
        if getattr(method, "__isabstractmethod__", False):
            setattr(cls, name, patch_grpc_method(name, fallback))
        elif name[0].isupper() and inspect.isfunction(method):
            setattr(cls, name, patch_grpc_method(name, method))

    cls.__abstractmethods__ = frozenset()
    return cls


class ResponseNotConsumed(Exception):
    def __init__(self, unconsumed_requests: list[str]):
        self.unconsumed_requests = unconsumed_requests
        request_count = Counter(unconsumed_requests)
        super().__init__(f"Expected but did not receive the following requests: {request_count}")


class InterceptionContext:
    def __init__(self, servicer):
        self._servicer = servicer
        self.calls: list[tuple[str, Any]] = []  # List[Tuple[method_name, message]]
        self.custom_responses: dict[str, list[tuple[Callable[[Any], bool], list[Any]]]] = defaultdict(list)
        self.custom_defaults: dict[str, Callable[["MockClientServicer", grpclib.server.Stream], Awaitable[None]]] = {}

    def add_response(
        self, method_name: str, first_payload, *, request_filter: Callable[[Any], bool] = lambda req: True
    ):
        """Adds one response payload to an expected queue of responses for a method.

        These responses will be used once each instead of calling the MockServicer's
        implementation of the method.

        The interception context will throw an exception on exit if not all of the added
        responses have been consumed.
        """
        self.custom_responses[method_name].append((request_filter, [first_payload]))

    def set_responder(
        self, method_name: str, responder: Callable[["MockClientServicer", grpclib.server.Stream], Awaitable[None]]
    ):
        """Replace the default responder from the MockClientServicer with a custom implementation

        ```python notest
        def custom_responder(servicer, stream):
            request = stream.recv_message()
            await stream.send_message(api_pb2.SomeMethodResponse(foo=123))

        with servicer.intercept() as ctx:
            ctx.set_responder("SomeMethod", custom_responder)
        ```

        Responses added via `.add_response()` take precedence over the use of this replacement
        """
        self.custom_defaults[method_name] = responder

    def pop_request(self, method_name):
        # fast forward to the next request of type method_name
        # dropping any preceding requests if there is a match
        # returns the payload of the request
        for i, (_method_name, msg) in enumerate(self.calls):
            if _method_name == method_name:
                self.calls = self.calls[i + 1 :]
                return msg

        raise KeyError(f"No message of that type in call list: {self.calls}")

    def get_requests(self, method_name: str) -> list[Any]:
        if not hasattr(self._servicer, method_name):
            # we check this to prevent things like `assert ctx.get_requests("ASdfFunctionCreate") == 0` passing
            raise ValueError(f"{method_name} not in MockServicer - did you spell it right?")
        return [msg for _method_name, msg in self.calls if _method_name == method_name]

    def _add_recv(self, method_name: str, msg):
        self.calls.append((method_name, msg))

    def _next_custom_responder(self, method_name, request):
        method_responses = self.custom_responses[method_name]
        for i, (request_filter, response_messages) in enumerate(method_responses):
            try:
                request_matches = request_filter(request)
            except Exception:
                logging.exception("Error when filtering requests")
                raise

            if request_matches:
                next_response_messages = response_messages
                self.custom_responses[method_name] = method_responses[:i] + method_responses[i + 1 :]
                break
        else:
            custom_default = self.custom_defaults.get(method_name)
            if not custom_default:
                return None
            return custom_default

        # build a new temporary responder based on the next queued response messages (added via add_response)
        async def responder(servicer_self, stream):
            await stream.recv_message()  # get the input message so we can track that
            for msg in next_response_messages:
                await stream.send_message(msg)

        return responder

    def _assert_responses_consumed(self):
        unconsumed = []
        for method_name, queued_responses in self.custom_responses.items():
            unconsumed += [method_name] * len(queued_responses)

        if unconsumed:
            raise ResponseNotConsumed(unconsumed)


class InterceptedStream:
    def __init__(self, interception_context: InterceptionContext, method_name: str, stream):
        self.interception_context = interception_context
        self.method_name = method_name
        self.stream = stream
        self.request_message = None

    async def initialize(self):
        self.request_message = await self.recv_message()
        return self

    async def recv_message(self):
        if self.request_message:
            ret = self.request_message
            self.request_message = None
            return ret

        msg = await self.stream.recv_message()
        self.interception_context._add_recv(self.method_name, msg)
        return msg

    async def send_message(self, msg):
        await self.stream.send_message(msg)

    def __getattr__(self, attr):
        return getattr(self.stream, attr)



# Copyright Modal Labs 2023
import itertools
import os
import webbrowser
from collections.abc import AsyncGenerator
from typing import Optional

import aiohttp.web
from rich.console import Console
from synchronicity.async_wrap import asynccontextmanager

from modal_proto import api_pb2

from ._utils.async_utils import synchronize_api
from ._utils.http_utils import run_temporary_http_server
from .client import _Client
from .config import _lookup_workspace, _store_user_config, config, config_profiles, user_config_path
from .exception import AuthError


class _TokenFlow:
    def __init__(self, client: _Client):
        self.stub = client.stub

    @asynccontextmanager
    async def start(
        self, utm_source: Optional[str] = None, next_url: Optional[str] = None
    ) -> AsyncGenerator[tuple[str, str, str], None]:
        """mdmd:hidden"""
        # Run a temporary http server returning the token id on /
        # This helps us add direct validation later
        # TODO(erikbern): handle failure launching server

        async def slash(request):
            headers = {"Access-Control-Allow-Origin": "*"}
            return aiohttp.web.Response(text=self.token_flow_id, headers=headers)

        app = aiohttp.web.Application()
        app.add_routes([aiohttp.web.get("/", slash)])
        async with run_temporary_http_server(app) as url:
            req = api_pb2.TokenFlowCreateRequest(
                utm_source=utm_source,
                next_url=next_url,
                localhost_port=int(url.split(":")[-1]),
            )
            resp = await self.stub.TokenFlowCreate(req)
            self.token_flow_id = resp.token_flow_id
            self.wait_secret = resp.wait_secret
            yield (resp.token_flow_id, resp.web_url, resp.code)

    async def finish(
        self, timeout: float = 40.0, grpc_extra_timeout: float = 5.0
    ) -> Optional[api_pb2.TokenFlowWaitResponse]:
        """mdmd:hidden"""
        # Wait for token flow to finish
        req = api_pb2.TokenFlowWaitRequest(
            token_flow_id=self.token_flow_id, timeout=timeout, wait_secret=self.wait_secret
        )
        resp = await self.stub.TokenFlowWait(req, timeout=(timeout + grpc_extra_timeout))
        if not resp.timeout:
            return resp
        else:
            return None


TokenFlow = synchronize_api(_TokenFlow)


async def _new_token(
    *,
    profile: Optional[str] = None,
    activate: bool = True,
    verify: bool = True,
    source: Optional[str] = None,
    next_url: Optional[str] = None,
):
    server_url = config.get("server_url", profile=profile)

    console = Console()

    result: Optional[api_pb2.TokenFlowWaitResponse] = None
    async with _Client.anonymous(server_url) as client:
        token_flow = _TokenFlow(client)

        async with token_flow.start(source, next_url) as (_, web_url, code):
            with console.status("Waiting for authentication in the web browser", spinner="dots"):
                # Open the web url in the browser
                if _open_url(web_url):
                    console.print(
                        "The web browser should have opened for you to authenticate and get an API token.\n"
                        "If it didn't, please copy this URL into your web browser manually:\n"
                    )
                else:
                    console.print(
                        "[red]Was not able to launch web browser[/red]\n"
                        "Please go to this URL manually and complete the flow:\n"
                    )
                console.print(f"[link={web_url}]{web_url}[/link]\n")
                if code:
                    console.print(f"Enter this code: [yellow]{code}[/yellow]\n")

            with console.status("Waiting for token flow to complete...", spinner="dots") as status:
                for attempt in itertools.count():
                    result = await token_flow.finish()
                    if result is not None:
                        break
                    status.update(f"Waiting for token flow to complete... (attempt {attempt + 2})")

        console.print("[green]Web authentication finished successfully![/green]")

    assert result is not None

    if result.workspace_username:
        console.print(
            f"[green]Token is connected to the [magenta]{result.workspace_username}[/magenta] workspace.[/green]"
        )

    await _set_token(result.token_id, result.token_secret, profile=profile, activate=activate, verify=verify)


async def _set_token(
    token_id: str,
    token_secret: str,
    *,
    profile: Optional[str] = None,
    activate: bool = True,
    verify: bool = True,
):
    # TODO add server_url as a parameter for verification?
    server_url = config.get("server_url", profile=profile)
    console = Console()
    if verify:
        console.print(f"Verifying token against [blue]{server_url}[/blue]")
        await _Client.verify(server_url, (token_id, token_secret))
        console.print("[green]Token verified successfully![/green]")

    if profile is None:
        if "MODAL_PROFILE" in os.environ:
            profile = os.environ["MODAL_PROFILE"]
        else:
            try:
                workspace = await _lookup_workspace(server_url, token_id, token_secret)
            except AuthError as exc:
                if not verify:
                    # Improve the error message for verification failure with --no-verify to reduce surprise
                    msg = "No profile name given, but could not authenticate client to look up workspace name."
                    raise AuthError(msg) from exc
                raise exc
            profile = workspace.username

    config_data = {"token_id": token_id, "token_secret": token_secret}
    # Activate the profile when requested or if no other profiles currently exist
    active_profile = profile if (activate or not config_profiles()) else None
    with console.status("Storing token", spinner="dots"):
        _store_user_config(config_data, profile=profile, active_profile=active_profile)
    console.print(
        f"[green]Token written to [magenta]{user_config_path}[/magenta] in profile "
        f"[magenta]{profile}[/magenta].[/green]"
    )


def _open_url(url: str) -> bool:
    """Opens url in web browser, making sure we use a modern one (not Lynx etc)"""
    if "PYTEST_CURRENT_TEST" in os.environ:
        return False
    try:
        browser = webbrowser.get()
        # zpresto defines `BROWSER=open` by default on macOS, which causes `webbrowser` to return `GenericBrowser`.
        if isinstance(browser, webbrowser.GenericBrowser) and browser.name != "open":
            return False
        else:
            return browser.open_new_tab(url)
    except webbrowser.Error:
        return False



# Copyright Modal Labs 2023
"""Utilities to generate random valid Protobuf messages for testing.

This is based on https://github.com/yupingso/randomproto but customizable for
Modal, with random seeds, and it supports oneofs, and Protobuf v4.
"""

import string
from random import Random
from typing import Any, Callable, Optional, TypeVar, Union

from google.protobuf.descriptor import Descriptor, FieldDescriptor

T = TypeVar("T")

_FIELD_RANDOM_GENERATOR: dict[int, Callable[[Random], Any]] = {
    FieldDescriptor.TYPE_DOUBLE: lambda rand: rand.normalvariate(0, 1),
    FieldDescriptor.TYPE_FLOAT: lambda rand: rand.normalvariate(0, 1),
    FieldDescriptor.TYPE_INT32: lambda rand: int.from_bytes(rand.randbytes(4), "little", signed=True),
    FieldDescriptor.TYPE_INT64: lambda rand: int.from_bytes(rand.randbytes(8), "little", signed=True),
    FieldDescriptor.TYPE_UINT32: lambda rand: int.from_bytes(rand.randbytes(4), "little"),
    FieldDescriptor.TYPE_UINT64: lambda rand: int.from_bytes(rand.randbytes(8), "little"),
    FieldDescriptor.TYPE_SINT32: lambda rand: int.from_bytes(rand.randbytes(4), "little", signed=True),
    FieldDescriptor.TYPE_SINT64: lambda rand: int.from_bytes(rand.randbytes(8), "little", signed=True),
    FieldDescriptor.TYPE_FIXED32: lambda rand: int.from_bytes(rand.randbytes(4), "little"),
    FieldDescriptor.TYPE_FIXED64: lambda rand: int.from_bytes(rand.randbytes(8), "little"),
    FieldDescriptor.TYPE_SFIXED32: lambda rand: int.from_bytes(rand.randbytes(4), "little", signed=True),
    FieldDescriptor.TYPE_SFIXED64: lambda rand: int.from_bytes(rand.randbytes(8), "little", signed=True),
    FieldDescriptor.TYPE_BOOL: lambda rand: rand.choice([True, False]),
    FieldDescriptor.TYPE_STRING: lambda rand: "".join(
        rand.choice(string.printable) for _ in range(int(rand.expovariate(0.15)))
    ),
    FieldDescriptor.TYPE_BYTES: lambda rand: rand.randbytes(int(rand.expovariate(0.15))),
}


def _fill(msg, desc: Descriptor, rand: Random) -> None:
    field: FieldDescriptor
    oneof_fields: set[str] = set()
    for oneof in desc.oneofs:
        oneof_field: Union[FieldDescriptor, None] = rand.choice(list(oneof.fields) + [None])
        if oneof_field is not None:
            oneof_fields.add(oneof_field.name)
    for field in desc.fields:
        if field.containing_oneof is not None and field.name not in oneof_fields:
            continue
        is_message = field.type == FieldDescriptor.TYPE_MESSAGE
        is_repeated = field.label == FieldDescriptor.LABEL_REPEATED
        if is_message:
            msg_field = getattr(msg, field.name)
            if is_repeated:
                num = rand.randint(0, 2)
                for _ in range(num):
                    element = msg_field.add()
                    _fill(element, field.message_type, rand)
            else:
                _fill(msg_field, field.message_type, rand)
        else:
            if field.type == FieldDescriptor.TYPE_ENUM:
                enum_values = [x.number for x in field.enum_type.values]
                generator = lambda rand: rand.choice(enum_values)  # noqa: E731

            else:
                generator = _FIELD_RANDOM_GENERATOR[field.type]
            if is_repeated:
                num = rand.randint(0, 2)
                msg_field = getattr(msg, field.name)
                for _ in range(num):
                    msg_field.append(generator(rand))
            else:
                setattr(msg, field.name, generator(rand))


def rand_pb(proto: type[T], rand: Optional[Random] = None) -> T:
    """Generate a pseudorandom protobuf message.

    ```python notest
    rand = random.Random(42)
    definition = rand_pb(api_pb2.Function, rand)
    ```
    """
    if rand is None:
        rand = Random(0)  # note: deterministic seed if not specified
    msg = proto()
    _fill(msg, proto.DESCRIPTOR, rand)  # type: ignore
    return msg



# Copyright Modal Labs 2022
import asyncio
import contextlib
import platform
import socket
import time
import typing
import urllib.parse
import uuid
from collections.abc import AsyncIterator
from dataclasses import dataclass
from typing import (
    Any,
    Optional,
    TypeVar,
)

import grpclib.client
import grpclib.config
import grpclib.events
import grpclib.protocol
import grpclib.stream
from google.protobuf.message import Message
from grpclib import GRPCError, Status
from grpclib.exceptions import StreamTerminatedError
from grpclib.protocol import H2Protocol

from modal.exception import AuthError, ConnectionError
from modal_version import __version__

from .logger import logger

RequestType = TypeVar("RequestType", bound=Message)
ResponseType = TypeVar("ResponseType", bound=Message)

if typing.TYPE_CHECKING:
    import modal.client

# Monkey patches grpclib to have a Modal User Agent header.
grpclib.client.USER_AGENT = "modal-client/{version} ({sys}; {py}/{py_ver})'".format(
    version=__version__,
    sys=platform.system(),
    py=platform.python_implementation(),
    py_ver=platform.python_version(),
).lower()


class Subchannel:
    protocol: H2Protocol
    created_at: float
    requests: int

    def __init__(self, protocol: H2Protocol) -> None:
        self.protocol = protocol
        self.created_at = time.time()
        self.requests = 0

    def connected(self):
        if hasattr(self.protocol.handler, "connection_lost"):
            # AbstractHandler doesn't have connection_lost, but Handler does
            return not self.protocol.handler.connection_lost  # type: ignore
        return True


RETRYABLE_GRPC_STATUS_CODES = [
    Status.DEADLINE_EXCEEDED,
    Status.UNAVAILABLE,
    Status.CANCELLED,
    Status.INTERNAL,
]

@dataclass
class RetryWarningMessage:
    message: str
    warning_interval: int
    errors_to_warn_for: typing.List[Status]

def create_channel(
    server_url: str,
    metadata: dict[str, str] = {},
) -> grpclib.client.Channel:
    """Creates a grpclib.Channel.

    Either to be used directly by a GRPC stub, or indirectly used through the channel pool.
    See `create_channel`.
    """
    o = urllib.parse.urlparse(server_url)

    channel: grpclib.client.Channel
    config = grpclib.config.Configuration(
        http2_connection_window_size=64 * 1024 * 1024,  # 64 MiB
        http2_stream_window_size=64 * 1024 * 1024,  # 64 MiB
    )

    if o.scheme == "unix":
        channel = grpclib.client.Channel(path=o.path, config=config)  # probably pointless to use a pool ever
    elif o.scheme in ("http", "https"):
        target = o.netloc
        parts = target.split(":")
        assert 1 <= len(parts) <= 2, "Invalid target location: " + target
        ssl = o.scheme.endswith("s")
        host = parts[0]
        port = int(parts[1]) if len(parts) == 2 else 443 if ssl else 80
        channel = grpclib.client.Channel(host, port, ssl=ssl, config=config)
    else:
        raise Exception(f"Unknown scheme: {o.scheme}")

    target = o.path if o.scheme == "unix" else o.netloc
    logger.debug(f"Connecting to {target} using scheme {o.scheme}")

    # Inject metadata for the client.
    async def send_request(event: grpclib.events.SendRequest) -> None:
        for k, v in metadata.items():
            event.metadata[k] = v

        logger.debug(f"Sending request to {event.method_name}")

    grpclib.events.listen(channel, grpclib.events.SendRequest, send_request)

    return channel


async def connect_channel(channel: grpclib.client.Channel):
    """Connects socket (potentially raising errors raising to connectivity."""
    await channel.__connect__()


if typing.TYPE_CHECKING:
    import modal.client


async def unary_stream(
    method: "modal.client.UnaryStreamWrapper[RequestType, ResponseType]",
    request: RequestType,
    metadata: Optional[Any] = None,
) -> AsyncIterator[ResponseType]:
    # TODO: remove this, since we have a method now
    async for item in method.unary_stream(request, metadata):
        yield item


async def retry_transient_errors(
    fn: "modal.client.UnaryUnaryWrapper[RequestType, ResponseType]",
    *args,
    base_delay: float = 0.1,
    max_delay: float = 1,
    delay_factor: float = 2,
    max_retries: Optional[int] = 3,
    additional_status_codes: list = [],
    attempt_timeout: Optional[float] = None,  # timeout for each attempt
    total_timeout: Optional[float] = None,  # timeout for the entire function call
    attempt_timeout_floor=2.0,  # always have at least this much timeout (only for total_timeout)
    retry_warning_message: Optional[RetryWarningMessage] = None
) -> ResponseType:
    """Retry on transient gRPC failures with back-off until max_retries is reached.
    If max_retries is None, retry forever."""

    delay = base_delay
    n_retries = 0

    status_codes = [*RETRYABLE_GRPC_STATUS_CODES, *additional_status_codes]

    idempotency_key = str(uuid.uuid4())

    t0 = time.time()
    if total_timeout is not None:
        total_deadline = t0 + total_timeout
    else:
        total_deadline = None

    while True:
        metadata = [("x-idempotency-key", idempotency_key), ("x-retry-attempt", str(n_retries))]
        if n_retries > 0:
            metadata.append(("x-retry-delay", str(time.time() - t0)))
        timeouts = []
        if attempt_timeout is not None:
            timeouts.append(attempt_timeout)
        if total_timeout is not None:
            timeouts.append(max(total_deadline - time.time(), attempt_timeout_floor))
        if timeouts:
            timeout = min(timeouts)  # In case the function provided both types of timeouts
        else:
            timeout = None
        try:
            return await fn(*args, metadata=metadata, timeout=timeout)
        except (StreamTerminatedError, GRPCError, OSError, asyncio.TimeoutError, AttributeError) as exc:
            if isinstance(exc, GRPCError) and exc.status not in status_codes:
                if exc.status == Status.UNAUTHENTICATED:
                    raise AuthError(exc.message)
                else:
                    raise exc

            if max_retries is not None and n_retries >= max_retries:
                final_attempt = True
            elif total_deadline is not None and time.time() + delay + attempt_timeout_floor >= total_deadline:
                final_attempt = True
            else:
                final_attempt = False

            if final_attempt:
                logger.debug(
                    f"Final attempt failed with {repr(exc)} {n_retries=} {delay=} "
                    f"{total_deadline=} for {fn.name} ({idempotency_key[:8]})"
                )
                if isinstance(exc, OSError):
                    raise ConnectionError(str(exc))
                elif isinstance(exc, asyncio.TimeoutError):
                    raise ConnectionError(str(exc))
                else:
                    raise exc

            if isinstance(exc, AttributeError) and "_write_appdata" not in str(exc):
                # StreamTerminatedError are not properly raised in grpclib<=0.4.7
                # fixed in https://github.com/vmagamedov/grpclib/issues/185
                # TODO: update to newer version (>=0.4.8) once stable
                raise exc

            logger.debug(f"Retryable failure {repr(exc)} {n_retries=} {delay=} for {fn.name} ({idempotency_key[:8]})")

            n_retries += 1

            if (
                retry_warning_message
                and n_retries % retry_warning_message.warning_interval == 0
                and isinstance(exc, GRPCError)
                and exc.status in retry_warning_message.errors_to_warn_for
            ):
                logger.warning(retry_warning_message.message)

            await asyncio.sleep(delay)
            delay = min(delay * delay_factor, max_delay)


def find_free_port() -> int:
    """
    Find a free TCP port, useful for testing.

    WARN: if a returned free port is not bound immediately by the caller, that same port
    may be returned in subsequent calls to this function, potentially creating port collisions.
    """
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def get_proto_oneof(message: Message, oneof_group: str) -> Optional[Message]:
    oneof_field = message.WhichOneof(oneof_group)
    if oneof_field is None:
        return None

    return getattr(message, oneof_field)



import google.protobuf.message
import modal._object
import modal.client
import modal.object
import modal_proto.api_pb2
import typing
import typing_extensions

class EnvironmentSettings:
    image_builder_version: str
    webhook_suffix: str

    def __init__(self, image_builder_version: str, webhook_suffix: str) -> None: ...
    def __repr__(self): ...
    def __eq__(self, other): ...
    def __setattr__(self, name, value): ...
    def __delattr__(self, name): ...
    def __hash__(self): ...

class _Environment(modal._object._Object):
    _settings: EnvironmentSettings

    def __init__(self): ...
    def _hydrate_metadata(self, metadata: google.protobuf.message.Message): ...
    @staticmethod
    def from_name(name: str, create_if_missing: bool = False): ...
    @staticmethod
    async def lookup(
        name: str, client: typing.Optional[modal.client._Client] = None, create_if_missing: bool = False
    ): ...

class Environment(modal.object.Object):
    _settings: EnvironmentSettings

    def __init__(self): ...
    def _hydrate_metadata(self, metadata: google.protobuf.message.Message): ...
    @staticmethod
    def from_name(name: str, create_if_missing: bool = False): ...

    class __lookup_spec(typing_extensions.Protocol):
        def __call__(
            self, name: str, client: typing.Optional[modal.client.Client] = None, create_if_missing: bool = False
        ): ...
        async def aio(
            self, name: str, client: typing.Optional[modal.client.Client] = None, create_if_missing: bool = False
        ): ...

    lookup: __lookup_spec

async def _get_environment_cached(name: str, client: modal.client._Client) -> _Environment: ...

class __delete_environment_spec(typing_extensions.Protocol):
    def __call__(self, name: str, client: typing.Optional[modal.client.Client] = None): ...
    async def aio(self, name: str, client: typing.Optional[modal.client.Client] = None): ...

delete_environment: __delete_environment_spec

class __update_environment_spec(typing_extensions.Protocol):
    def __call__(
        self,
        current_name: str,
        *,
        new_name: typing.Optional[str] = None,
        new_web_suffix: typing.Optional[str] = None,
        client: typing.Optional[modal.client.Client] = None,
    ): ...
    async def aio(
        self,
        current_name: str,
        *,
        new_name: typing.Optional[str] = None,
        new_web_suffix: typing.Optional[str] = None,
        client: typing.Optional[modal.client.Client] = None,
    ): ...

update_environment: __update_environment_spec

class __create_environment_spec(typing_extensions.Protocol):
    def __call__(self, name: str, client: typing.Optional[modal.client.Client] = None): ...
    async def aio(self, name: str, client: typing.Optional[modal.client.Client] = None): ...

create_environment: __create_environment_spec

class __list_environments_spec(typing_extensions.Protocol):
    def __call__(
        self, client: typing.Optional[modal.client.Client] = None
    ) -> list[modal_proto.api_pb2.EnvironmentListItem]: ...
    async def aio(
        self, client: typing.Optional[modal.client.Client] = None
    ) -> list[modal_proto.api_pb2.EnvironmentListItem]: ...

list_environments: __list_environments_spec

def ensure_env(environment_name: typing.Optional[str] = None) -> str: ...

ENVIRONMENT_CACHE: dict[str, _Environment]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import collections.abc
import google.protobuf.empty_pb2
import grpc
import modal_proto.api_pb2

class ModalClientStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    AppClientDisconnect: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppClientDisconnectRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Apps"""
    AppCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppCreateRequest,
        modal_proto.api_pb2.AppCreateResponse,
    ]
    AppDeploy: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppDeployRequest,
        modal_proto.api_pb2.AppDeployResponse,
    ]
    AppDeploymentHistory: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppDeploymentHistoryRequest,
        modal_proto.api_pb2.AppDeploymentHistoryResponse,
    ]
    AppGetByDeploymentName: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppGetByDeploymentNameRequest,
        modal_proto.api_pb2.AppGetByDeploymentNameResponse,
    ]
    AppGetLayout: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppGetLayoutRequest,
        modal_proto.api_pb2.AppGetLayoutResponse,
    ]
    AppGetLogs: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.AppGetLogsRequest,
        modal_proto.api_pb2.TaskLogsBatch,
    ]
    AppGetObjects: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppGetObjectsRequest,
        modal_proto.api_pb2.AppGetObjectsResponse,
    ]
    AppGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppGetOrCreateRequest,
        modal_proto.api_pb2.AppGetOrCreateResponse,
    ]
    AppHeartbeat: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppHeartbeatRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    AppList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppListRequest,
        modal_proto.api_pb2.AppListResponse,
    ]
    AppLookup: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppLookupRequest,
        modal_proto.api_pb2.AppLookupResponse,
    ]
    AppPublish: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppPublishRequest,
        modal_proto.api_pb2.AppPublishResponse,
    ]
    AppRollback: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppRollbackRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    AppSetObjects: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppSetObjectsRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    AppStop: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.AppStopRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    BlobCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.BlobCreateRequest,
        modal_proto.api_pb2.BlobCreateResponse,
    ]
    """Blobs"""
    BlobGet: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.BlobGetRequest,
        modal_proto.api_pb2.BlobGetResponse,
    ]
    ClassCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ClassCreateRequest,
        modal_proto.api_pb2.ClassCreateResponse,
    ]
    """Classes"""
    ClassGet: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ClassGetRequest,
        modal_proto.api_pb2.ClassGetResponse,
    ]
    ClientHello: grpc.UnaryUnaryMultiCallable[
        google.protobuf.empty_pb2.Empty,
        modal_proto.api_pb2.ClientHelloResponse,
    ]
    """Clients"""
    ContainerCheckpoint: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ContainerCheckpointRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Container"""
    ContainerExec: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ContainerExecRequest,
        modal_proto.api_pb2.ContainerExecResponse,
    ]
    ContainerExecGetOutput: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.ContainerExecGetOutputRequest,
        modal_proto.api_pb2.RuntimeOutputBatch,
    ]
    ContainerExecPutInput: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ContainerExecPutInputRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    ContainerExecWait: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ContainerExecWaitRequest,
        modal_proto.api_pb2.ContainerExecWaitResponse,
    ]
    ContainerFilesystemExec: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ContainerFilesystemExecRequest,
        modal_proto.api_pb2.ContainerFilesystemExecResponse,
    ]
    ContainerFilesystemExecGetOutput: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.ContainerFilesystemExecGetOutputRequest,
        modal_proto.api_pb2.FilesystemRuntimeOutputBatch,
    ]
    ContainerHeartbeat: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ContainerHeartbeatRequest,
        modal_proto.api_pb2.ContainerHeartbeatResponse,
    ]
    ContainerHello: grpc.UnaryUnaryMultiCallable[
        google.protobuf.empty_pb2.Empty,
        google.protobuf.empty_pb2.Empty,
    ]
    ContainerLog: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ContainerLogRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    ContainerStop: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ContainerStopRequest,
        modal_proto.api_pb2.ContainerStopResponse,
    ]
    DictClear: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictClearRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Dicts"""
    DictContains: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictContainsRequest,
        modal_proto.api_pb2.DictContainsResponse,
    ]
    DictContents: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.DictContentsRequest,
        modal_proto.api_pb2.DictEntry,
    ]
    DictDelete: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictDeleteRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    DictGet: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictGetRequest,
        modal_proto.api_pb2.DictGetResponse,
    ]
    DictGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictGetOrCreateRequest,
        modal_proto.api_pb2.DictGetOrCreateResponse,
    ]
    DictHeartbeat: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictHeartbeatRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    DictLen: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictLenRequest,
        modal_proto.api_pb2.DictLenResponse,
    ]
    DictList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictListRequest,
        modal_proto.api_pb2.DictListResponse,
    ]
    DictPop: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictPopRequest,
        modal_proto.api_pb2.DictPopResponse,
    ]
    DictUpdate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DictUpdateRequest,
        modal_proto.api_pb2.DictUpdateResponse,
    ]
    DomainCertificateVerify: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DomainCertificateVerifyRequest,
        modal_proto.api_pb2.DomainCertificateVerifyResponse,
    ]
    """Domains"""
    DomainCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DomainCreateRequest,
        modal_proto.api_pb2.DomainCreateResponse,
    ]
    DomainList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.DomainListRequest,
        modal_proto.api_pb2.DomainListResponse,
    ]
    EnvironmentCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.EnvironmentCreateRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Environments"""
    EnvironmentDelete: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.EnvironmentDeleteRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    EnvironmentGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.EnvironmentGetOrCreateRequest,
        modal_proto.api_pb2.EnvironmentGetOrCreateResponse,
    ]
    EnvironmentList: grpc.UnaryUnaryMultiCallable[
        google.protobuf.empty_pb2.Empty,
        modal_proto.api_pb2.EnvironmentListResponse,
    ]
    EnvironmentUpdate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.EnvironmentUpdateRequest,
        modal_proto.api_pb2.EnvironmentListItem,
    ]
    FunctionAsyncInvoke: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionAsyncInvokeRequest,
        modal_proto.api_pb2.FunctionAsyncInvokeResponse,
    ]
    """Functions"""
    FunctionBindParams: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionBindParamsRequest,
        modal_proto.api_pb2.FunctionBindParamsResponse,
    ]
    FunctionCallCancel: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionCallCancelRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    FunctionCallGetDataIn: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.FunctionCallGetDataRequest,
        modal_proto.api_pb2.DataChunk,
    ]
    FunctionCallGetDataOut: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.FunctionCallGetDataRequest,
        modal_proto.api_pb2.DataChunk,
    ]
    FunctionCallList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionCallListRequest,
        modal_proto.api_pb2.FunctionCallListResponse,
    ]
    FunctionCallPutDataOut: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionCallPutDataRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    FunctionCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionCreateRequest,
        modal_proto.api_pb2.FunctionCreateResponse,
    ]
    FunctionGet: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionGetRequest,
        modal_proto.api_pb2.FunctionGetResponse,
    ]
    FunctionGetCallGraph: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionGetCallGraphRequest,
        modal_proto.api_pb2.FunctionGetCallGraphResponse,
    ]
    FunctionGetCurrentStats: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionGetCurrentStatsRequest,
        modal_proto.api_pb2.FunctionStats,
    ]
    FunctionGetDynamicConcurrency: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionGetDynamicConcurrencyRequest,
        modal_proto.api_pb2.FunctionGetDynamicConcurrencyResponse,
    ]
    FunctionGetInputs: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionGetInputsRequest,
        modal_proto.api_pb2.FunctionGetInputsResponse,
    ]
    """For containers to request next call"""
    FunctionGetOutputs: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionGetOutputsRequest,
        modal_proto.api_pb2.FunctionGetOutputsResponse,
    ]
    """Returns the next result(s) for an entire function call (FunctionMap)"""
    FunctionGetSerialized: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionGetSerializedRequest,
        modal_proto.api_pb2.FunctionGetSerializedResponse,
    ]
    FunctionMap: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionMapRequest,
        modal_proto.api_pb2.FunctionMapResponse,
    ]
    FunctionPrecreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionPrecreateRequest,
        modal_proto.api_pb2.FunctionPrecreateResponse,
    ]
    FunctionPutInputs: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionPutInputsRequest,
        modal_proto.api_pb2.FunctionPutInputsResponse,
    ]
    FunctionPutOutputs: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionPutOutputsRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """For containers to return result"""
    FunctionRetryInputs: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionRetryInputsRequest,
        modal_proto.api_pb2.FunctionRetryInputsResponse,
    ]
    FunctionStartPtyShell: grpc.UnaryUnaryMultiCallable[
        google.protobuf.empty_pb2.Empty,
        google.protobuf.empty_pb2.Empty,
    ]
    FunctionUpdateSchedulingParams: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.FunctionUpdateSchedulingParamsRequest,
        modal_proto.api_pb2.FunctionUpdateSchedulingParamsResponse,
    ]
    ImageFromId: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ImageFromIdRequest,
        modal_proto.api_pb2.ImageFromIdResponse,
    ]
    """Images"""
    ImageGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ImageGetOrCreateRequest,
        modal_proto.api_pb2.ImageGetOrCreateResponse,
    ]
    ImageJoinStreaming: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.ImageJoinStreamingRequest,
        modal_proto.api_pb2.ImageJoinStreamingResponse,
    ]
    MountGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.MountGetOrCreateRequest,
        modal_proto.api_pb2.MountGetOrCreateResponse,
    ]
    """Mounts"""
    MountPutFile: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.MountPutFileRequest,
        modal_proto.api_pb2.MountPutFileResponse,
    ]
    NotebookKernelPublishResults: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.NotebookKernelPublishResultsRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Notebooks"""
    ProxyAddIp: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ProxyAddIpRequest,
        modal_proto.api_pb2.ProxyAddIpResponse,
    ]
    """Proxies"""
    ProxyCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ProxyCreateRequest,
        modal_proto.api_pb2.ProxyCreateResponse,
    ]
    ProxyDelete: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ProxyDeleteRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    ProxyGet: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ProxyGetRequest,
        modal_proto.api_pb2.ProxyGetResponse,
    ]
    ProxyGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ProxyGetOrCreateRequest,
        modal_proto.api_pb2.ProxyGetOrCreateResponse,
    ]
    ProxyList: grpc.UnaryUnaryMultiCallable[
        google.protobuf.empty_pb2.Empty,
        modal_proto.api_pb2.ProxyListResponse,
    ]
    ProxyRemoveIp: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.ProxyRemoveIpRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    QueueClear: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.QueueClearRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Queues"""
    QueueDelete: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.QueueDeleteRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    QueueGet: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.QueueGetRequest,
        modal_proto.api_pb2.QueueGetResponse,
    ]
    QueueGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.QueueGetOrCreateRequest,
        modal_proto.api_pb2.QueueGetOrCreateResponse,
    ]
    QueueHeartbeat: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.QueueHeartbeatRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    QueueLen: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.QueueLenRequest,
        modal_proto.api_pb2.QueueLenResponse,
    ]
    QueueList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.QueueListRequest,
        modal_proto.api_pb2.QueueListResponse,
    ]
    QueueNextItems: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.QueueNextItemsRequest,
        modal_proto.api_pb2.QueueNextItemsResponse,
    ]
    QueuePut: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.QueuePutRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    SandboxCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxCreateRequest,
        modal_proto.api_pb2.SandboxCreateResponse,
    ]
    """Sandboxes"""
    SandboxGetLogs: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.SandboxGetLogsRequest,
        modal_proto.api_pb2.TaskLogsBatch,
    ]
    SandboxGetTaskId: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxGetTaskIdRequest,
        modal_proto.api_pb2.SandboxGetTaskIdResponse,
    ]
    """needed for modal container exec"""
    SandboxGetTunnels: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxGetTunnelsRequest,
        modal_proto.api_pb2.SandboxGetTunnelsResponse,
    ]
    SandboxList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxListRequest,
        modal_proto.api_pb2.SandboxListResponse,
    ]
    SandboxRestore: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxRestoreRequest,
        modal_proto.api_pb2.SandboxRestoreResponse,
    ]
    SandboxSnapshot: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxSnapshotRequest,
        modal_proto.api_pb2.SandboxSnapshotResponse,
    ]
    SandboxSnapshotFs: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxSnapshotFsRequest,
        modal_proto.api_pb2.SandboxSnapshotFsResponse,
    ]
    SandboxSnapshotGet: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxSnapshotGetRequest,
        modal_proto.api_pb2.SandboxSnapshotGetResponse,
    ]
    SandboxSnapshotWait: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxSnapshotWaitRequest,
        modal_proto.api_pb2.SandboxSnapshotWaitResponse,
    ]
    SandboxStdinWrite: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxStdinWriteRequest,
        modal_proto.api_pb2.SandboxStdinWriteResponse,
    ]
    SandboxTagsSet: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxTagsSetRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    SandboxTerminate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxTerminateRequest,
        modal_proto.api_pb2.SandboxTerminateResponse,
    ]
    SandboxWait: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SandboxWaitRequest,
        modal_proto.api_pb2.SandboxWaitResponse,
    ]
    SecretDelete: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SecretDeleteRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Secrets"""
    SecretGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SecretGetOrCreateRequest,
        modal_proto.api_pb2.SecretGetOrCreateResponse,
    ]
    SecretList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SecretListRequest,
        modal_proto.api_pb2.SecretListResponse,
    ]
    SharedVolumeDelete: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SharedVolumeDeleteRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """SharedVolumes"""
    SharedVolumeGetFile: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SharedVolumeGetFileRequest,
        modal_proto.api_pb2.SharedVolumeGetFileResponse,
    ]
    SharedVolumeGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SharedVolumeGetOrCreateRequest,
        modal_proto.api_pb2.SharedVolumeGetOrCreateResponse,
    ]
    SharedVolumeHeartbeat: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SharedVolumeHeartbeatRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    SharedVolumeList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SharedVolumeListRequest,
        modal_proto.api_pb2.SharedVolumeListResponse,
    ]
    SharedVolumeListFiles: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SharedVolumeListFilesRequest,
        modal_proto.api_pb2.SharedVolumeListFilesResponse,
    ]
    SharedVolumeListFilesStream: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.SharedVolumeListFilesRequest,
        modal_proto.api_pb2.SharedVolumeListFilesResponse,
    ]
    SharedVolumePutFile: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SharedVolumePutFileRequest,
        modal_proto.api_pb2.SharedVolumePutFileResponse,
    ]
    SharedVolumeRemoveFile: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.SharedVolumeRemoveFileRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    TaskClusterHello: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.TaskClusterHelloRequest,
        modal_proto.api_pb2.TaskClusterHelloResponse,
    ]
    """Tasks"""
    TaskCurrentInputs: grpc.UnaryUnaryMultiCallable[
        google.protobuf.empty_pb2.Empty,
        modal_proto.api_pb2.TaskCurrentInputsResponse,
    ]
    TaskList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.TaskListRequest,
        modal_proto.api_pb2.TaskListResponse,
    ]
    TaskResult: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.TaskResultRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    TokenFlowCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.TokenFlowCreateRequest,
        modal_proto.api_pb2.TokenFlowCreateResponse,
    ]
    """Tokens (web auth flow)"""
    TokenFlowWait: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.TokenFlowWaitRequest,
        modal_proto.api_pb2.TokenFlowWaitResponse,
    ]
    TunnelStart: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.TunnelStartRequest,
        modal_proto.api_pb2.TunnelStartResponse,
    ]
    """Tunnels"""
    TunnelStop: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.TunnelStopRequest,
        modal_proto.api_pb2.TunnelStopResponse,
    ]
    VolumeCommit: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeCommitRequest,
        modal_proto.api_pb2.VolumeCommitResponse,
    ]
    """Volumes"""
    VolumeCopyFiles: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeCopyFilesRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    VolumeDelete: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeDeleteRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    VolumeGetFile: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeGetFileRequest,
        modal_proto.api_pb2.VolumeGetFileResponse,
    ]
    VolumeGetOrCreate: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeGetOrCreateRequest,
        modal_proto.api_pb2.VolumeGetOrCreateResponse,
    ]
    VolumeHeartbeat: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeHeartbeatRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    VolumeList: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeListRequest,
        modal_proto.api_pb2.VolumeListResponse,
    ]
    VolumeListFiles: grpc.UnaryStreamMultiCallable[
        modal_proto.api_pb2.VolumeListFilesRequest,
        modal_proto.api_pb2.VolumeListFilesResponse,
    ]
    VolumePutFiles: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumePutFilesRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    VolumeReload: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeReloadRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    VolumeRemoveFile: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeRemoveFileRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    VolumeRename: grpc.UnaryUnaryMultiCallable[
        modal_proto.api_pb2.VolumeRenameRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    WorkspaceNameLookup: grpc.UnaryUnaryMultiCallable[
        google.protobuf.empty_pb2.Empty,
        modal_proto.api_pb2.WorkspaceNameLookupResponse,
    ]
    """Workspaces"""

class ModalClientServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def AppClientDisconnect(
        self,
        request: modal_proto.api_pb2.AppClientDisconnectRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Apps"""
    @abc.abstractmethod
    def AppCreate(
        self,
        request: modal_proto.api_pb2.AppCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppCreateResponse: ...
    @abc.abstractmethod
    def AppDeploy(
        self,
        request: modal_proto.api_pb2.AppDeployRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppDeployResponse: ...
    @abc.abstractmethod
    def AppDeploymentHistory(
        self,
        request: modal_proto.api_pb2.AppDeploymentHistoryRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppDeploymentHistoryResponse: ...
    @abc.abstractmethod
    def AppGetByDeploymentName(
        self,
        request: modal_proto.api_pb2.AppGetByDeploymentNameRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppGetByDeploymentNameResponse: ...
    @abc.abstractmethod
    def AppGetLayout(
        self,
        request: modal_proto.api_pb2.AppGetLayoutRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppGetLayoutResponse: ...
    @abc.abstractmethod
    def AppGetLogs(
        self,
        request: modal_proto.api_pb2.AppGetLogsRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.TaskLogsBatch]: ...
    @abc.abstractmethod
    def AppGetObjects(
        self,
        request: modal_proto.api_pb2.AppGetObjectsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppGetObjectsResponse: ...
    @abc.abstractmethod
    def AppGetOrCreate(
        self,
        request: modal_proto.api_pb2.AppGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppGetOrCreateResponse: ...
    @abc.abstractmethod
    def AppHeartbeat(
        self,
        request: modal_proto.api_pb2.AppHeartbeatRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def AppList(
        self,
        request: modal_proto.api_pb2.AppListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppListResponse: ...
    @abc.abstractmethod
    def AppLookup(
        self,
        request: modal_proto.api_pb2.AppLookupRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppLookupResponse: ...
    @abc.abstractmethod
    def AppPublish(
        self,
        request: modal_proto.api_pb2.AppPublishRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.AppPublishResponse: ...
    @abc.abstractmethod
    def AppRollback(
        self,
        request: modal_proto.api_pb2.AppRollbackRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def AppSetObjects(
        self,
        request: modal_proto.api_pb2.AppSetObjectsRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def AppStop(
        self,
        request: modal_proto.api_pb2.AppStopRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def BlobCreate(
        self,
        request: modal_proto.api_pb2.BlobCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.BlobCreateResponse:
        """Blobs"""
    @abc.abstractmethod
    def BlobGet(
        self,
        request: modal_proto.api_pb2.BlobGetRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.BlobGetResponse: ...
    @abc.abstractmethod
    def ClassCreate(
        self,
        request: modal_proto.api_pb2.ClassCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ClassCreateResponse:
        """Classes"""
    @abc.abstractmethod
    def ClassGet(
        self,
        request: modal_proto.api_pb2.ClassGetRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ClassGetResponse: ...
    @abc.abstractmethod
    def ClientHello(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ClientHelloResponse:
        """Clients"""
    @abc.abstractmethod
    def ContainerCheckpoint(
        self,
        request: modal_proto.api_pb2.ContainerCheckpointRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Container"""
    @abc.abstractmethod
    def ContainerExec(
        self,
        request: modal_proto.api_pb2.ContainerExecRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ContainerExecResponse: ...
    @abc.abstractmethod
    def ContainerExecGetOutput(
        self,
        request: modal_proto.api_pb2.ContainerExecGetOutputRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.RuntimeOutputBatch]: ...
    @abc.abstractmethod
    def ContainerExecPutInput(
        self,
        request: modal_proto.api_pb2.ContainerExecPutInputRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def ContainerExecWait(
        self,
        request: modal_proto.api_pb2.ContainerExecWaitRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ContainerExecWaitResponse: ...
    @abc.abstractmethod
    def ContainerFilesystemExec(
        self,
        request: modal_proto.api_pb2.ContainerFilesystemExecRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ContainerFilesystemExecResponse: ...
    @abc.abstractmethod
    def ContainerFilesystemExecGetOutput(
        self,
        request: modal_proto.api_pb2.ContainerFilesystemExecGetOutputRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.FilesystemRuntimeOutputBatch]: ...
    @abc.abstractmethod
    def ContainerHeartbeat(
        self,
        request: modal_proto.api_pb2.ContainerHeartbeatRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ContainerHeartbeatResponse: ...
    @abc.abstractmethod
    def ContainerHello(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def ContainerLog(
        self,
        request: modal_proto.api_pb2.ContainerLogRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def ContainerStop(
        self,
        request: modal_proto.api_pb2.ContainerStopRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ContainerStopResponse: ...
    @abc.abstractmethod
    def DictClear(
        self,
        request: modal_proto.api_pb2.DictClearRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Dicts"""
    @abc.abstractmethod
    def DictContains(
        self,
        request: modal_proto.api_pb2.DictContainsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DictContainsResponse: ...
    @abc.abstractmethod
    def DictContents(
        self,
        request: modal_proto.api_pb2.DictContentsRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.DictEntry]: ...
    @abc.abstractmethod
    def DictDelete(
        self,
        request: modal_proto.api_pb2.DictDeleteRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def DictGet(
        self,
        request: modal_proto.api_pb2.DictGetRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DictGetResponse: ...
    @abc.abstractmethod
    def DictGetOrCreate(
        self,
        request: modal_proto.api_pb2.DictGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DictGetOrCreateResponse: ...
    @abc.abstractmethod
    def DictHeartbeat(
        self,
        request: modal_proto.api_pb2.DictHeartbeatRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def DictLen(
        self,
        request: modal_proto.api_pb2.DictLenRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DictLenResponse: ...
    @abc.abstractmethod
    def DictList(
        self,
        request: modal_proto.api_pb2.DictListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DictListResponse: ...
    @abc.abstractmethod
    def DictPop(
        self,
        request: modal_proto.api_pb2.DictPopRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DictPopResponse: ...
    @abc.abstractmethod
    def DictUpdate(
        self,
        request: modal_proto.api_pb2.DictUpdateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DictUpdateResponse: ...
    @abc.abstractmethod
    def DomainCertificateVerify(
        self,
        request: modal_proto.api_pb2.DomainCertificateVerifyRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DomainCertificateVerifyResponse:
        """Domains"""
    @abc.abstractmethod
    def DomainCreate(
        self,
        request: modal_proto.api_pb2.DomainCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DomainCreateResponse: ...
    @abc.abstractmethod
    def DomainList(
        self,
        request: modal_proto.api_pb2.DomainListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.DomainListResponse: ...
    @abc.abstractmethod
    def EnvironmentCreate(
        self,
        request: modal_proto.api_pb2.EnvironmentCreateRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Environments"""
    @abc.abstractmethod
    def EnvironmentDelete(
        self,
        request: modal_proto.api_pb2.EnvironmentDeleteRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def EnvironmentGetOrCreate(
        self,
        request: modal_proto.api_pb2.EnvironmentGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.EnvironmentGetOrCreateResponse: ...
    @abc.abstractmethod
    def EnvironmentList(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.EnvironmentListResponse: ...
    @abc.abstractmethod
    def EnvironmentUpdate(
        self,
        request: modal_proto.api_pb2.EnvironmentUpdateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.EnvironmentListItem: ...
    @abc.abstractmethod
    def FunctionAsyncInvoke(
        self,
        request: modal_proto.api_pb2.FunctionAsyncInvokeRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionAsyncInvokeResponse:
        """Functions"""
    @abc.abstractmethod
    def FunctionBindParams(
        self,
        request: modal_proto.api_pb2.FunctionBindParamsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionBindParamsResponse: ...
    @abc.abstractmethod
    def FunctionCallCancel(
        self,
        request: modal_proto.api_pb2.FunctionCallCancelRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def FunctionCallGetDataIn(
        self,
        request: modal_proto.api_pb2.FunctionCallGetDataRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.DataChunk]: ...
    @abc.abstractmethod
    def FunctionCallGetDataOut(
        self,
        request: modal_proto.api_pb2.FunctionCallGetDataRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.DataChunk]: ...
    @abc.abstractmethod
    def FunctionCallList(
        self,
        request: modal_proto.api_pb2.FunctionCallListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionCallListResponse: ...
    @abc.abstractmethod
    def FunctionCallPutDataOut(
        self,
        request: modal_proto.api_pb2.FunctionCallPutDataRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def FunctionCreate(
        self,
        request: modal_proto.api_pb2.FunctionCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionCreateResponse: ...
    @abc.abstractmethod
    def FunctionGet(
        self,
        request: modal_proto.api_pb2.FunctionGetRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionGetResponse: ...
    @abc.abstractmethod
    def FunctionGetCallGraph(
        self,
        request: modal_proto.api_pb2.FunctionGetCallGraphRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionGetCallGraphResponse: ...
    @abc.abstractmethod
    def FunctionGetCurrentStats(
        self,
        request: modal_proto.api_pb2.FunctionGetCurrentStatsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionStats: ...
    @abc.abstractmethod
    def FunctionGetDynamicConcurrency(
        self,
        request: modal_proto.api_pb2.FunctionGetDynamicConcurrencyRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionGetDynamicConcurrencyResponse: ...
    @abc.abstractmethod
    def FunctionGetInputs(
        self,
        request: modal_proto.api_pb2.FunctionGetInputsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionGetInputsResponse:
        """For containers to request next call"""
    @abc.abstractmethod
    def FunctionGetOutputs(
        self,
        request: modal_proto.api_pb2.FunctionGetOutputsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionGetOutputsResponse:
        """Returns the next result(s) for an entire function call (FunctionMap)"""
    @abc.abstractmethod
    def FunctionGetSerialized(
        self,
        request: modal_proto.api_pb2.FunctionGetSerializedRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionGetSerializedResponse: ...
    @abc.abstractmethod
    def FunctionMap(
        self,
        request: modal_proto.api_pb2.FunctionMapRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionMapResponse: ...
    @abc.abstractmethod
    def FunctionPrecreate(
        self,
        request: modal_proto.api_pb2.FunctionPrecreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionPrecreateResponse: ...
    @abc.abstractmethod
    def FunctionPutInputs(
        self,
        request: modal_proto.api_pb2.FunctionPutInputsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionPutInputsResponse: ...
    @abc.abstractmethod
    def FunctionPutOutputs(
        self,
        request: modal_proto.api_pb2.FunctionPutOutputsRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """For containers to return result"""
    @abc.abstractmethod
    def FunctionRetryInputs(
        self,
        request: modal_proto.api_pb2.FunctionRetryInputsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionRetryInputsResponse: ...
    @abc.abstractmethod
    def FunctionStartPtyShell(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def FunctionUpdateSchedulingParams(
        self,
        request: modal_proto.api_pb2.FunctionUpdateSchedulingParamsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.FunctionUpdateSchedulingParamsResponse: ...
    @abc.abstractmethod
    def ImageFromId(
        self,
        request: modal_proto.api_pb2.ImageFromIdRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ImageFromIdResponse:
        """Images"""
    @abc.abstractmethod
    def ImageGetOrCreate(
        self,
        request: modal_proto.api_pb2.ImageGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ImageGetOrCreateResponse: ...
    @abc.abstractmethod
    def ImageJoinStreaming(
        self,
        request: modal_proto.api_pb2.ImageJoinStreamingRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.ImageJoinStreamingResponse]: ...
    @abc.abstractmethod
    def MountGetOrCreate(
        self,
        request: modal_proto.api_pb2.MountGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.MountGetOrCreateResponse:
        """Mounts"""
    @abc.abstractmethod
    def MountPutFile(
        self,
        request: modal_proto.api_pb2.MountPutFileRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.MountPutFileResponse: ...
    @abc.abstractmethod
    def NotebookKernelPublishResults(
        self,
        request: modal_proto.api_pb2.NotebookKernelPublishResultsRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Notebooks"""
    @abc.abstractmethod
    def ProxyAddIp(
        self,
        request: modal_proto.api_pb2.ProxyAddIpRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ProxyAddIpResponse:
        """Proxies"""
    @abc.abstractmethod
    def ProxyCreate(
        self,
        request: modal_proto.api_pb2.ProxyCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ProxyCreateResponse: ...
    @abc.abstractmethod
    def ProxyDelete(
        self,
        request: modal_proto.api_pb2.ProxyDeleteRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def ProxyGet(
        self,
        request: modal_proto.api_pb2.ProxyGetRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ProxyGetResponse: ...
    @abc.abstractmethod
    def ProxyGetOrCreate(
        self,
        request: modal_proto.api_pb2.ProxyGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ProxyGetOrCreateResponse: ...
    @abc.abstractmethod
    def ProxyList(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.ProxyListResponse: ...
    @abc.abstractmethod
    def ProxyRemoveIp(
        self,
        request: modal_proto.api_pb2.ProxyRemoveIpRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def QueueClear(
        self,
        request: modal_proto.api_pb2.QueueClearRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Queues"""
    @abc.abstractmethod
    def QueueDelete(
        self,
        request: modal_proto.api_pb2.QueueDeleteRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def QueueGet(
        self,
        request: modal_proto.api_pb2.QueueGetRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.QueueGetResponse: ...
    @abc.abstractmethod
    def QueueGetOrCreate(
        self,
        request: modal_proto.api_pb2.QueueGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.QueueGetOrCreateResponse: ...
    @abc.abstractmethod
    def QueueHeartbeat(
        self,
        request: modal_proto.api_pb2.QueueHeartbeatRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def QueueLen(
        self,
        request: modal_proto.api_pb2.QueueLenRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.QueueLenResponse: ...
    @abc.abstractmethod
    def QueueList(
        self,
        request: modal_proto.api_pb2.QueueListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.QueueListResponse: ...
    @abc.abstractmethod
    def QueueNextItems(
        self,
        request: modal_proto.api_pb2.QueueNextItemsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.QueueNextItemsResponse: ...
    @abc.abstractmethod
    def QueuePut(
        self,
        request: modal_proto.api_pb2.QueuePutRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def SandboxCreate(
        self,
        request: modal_proto.api_pb2.SandboxCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxCreateResponse:
        """Sandboxes"""
    @abc.abstractmethod
    def SandboxGetLogs(
        self,
        request: modal_proto.api_pb2.SandboxGetLogsRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.TaskLogsBatch]: ...
    @abc.abstractmethod
    def SandboxGetTaskId(
        self,
        request: modal_proto.api_pb2.SandboxGetTaskIdRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxGetTaskIdResponse:
        """needed for modal container exec"""
    @abc.abstractmethod
    def SandboxGetTunnels(
        self,
        request: modal_proto.api_pb2.SandboxGetTunnelsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxGetTunnelsResponse: ...
    @abc.abstractmethod
    def SandboxList(
        self,
        request: modal_proto.api_pb2.SandboxListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxListResponse: ...
    @abc.abstractmethod
    def SandboxRestore(
        self,
        request: modal_proto.api_pb2.SandboxRestoreRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxRestoreResponse: ...
    @abc.abstractmethod
    def SandboxSnapshot(
        self,
        request: modal_proto.api_pb2.SandboxSnapshotRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxSnapshotResponse: ...
    @abc.abstractmethod
    def SandboxSnapshotFs(
        self,
        request: modal_proto.api_pb2.SandboxSnapshotFsRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxSnapshotFsResponse: ...
    @abc.abstractmethod
    def SandboxSnapshotGet(
        self,
        request: modal_proto.api_pb2.SandboxSnapshotGetRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxSnapshotGetResponse: ...
    @abc.abstractmethod
    def SandboxSnapshotWait(
        self,
        request: modal_proto.api_pb2.SandboxSnapshotWaitRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxSnapshotWaitResponse: ...
    @abc.abstractmethod
    def SandboxStdinWrite(
        self,
        request: modal_proto.api_pb2.SandboxStdinWriteRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxStdinWriteResponse: ...
    @abc.abstractmethod
    def SandboxTagsSet(
        self,
        request: modal_proto.api_pb2.SandboxTagsSetRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def SandboxTerminate(
        self,
        request: modal_proto.api_pb2.SandboxTerminateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxTerminateResponse: ...
    @abc.abstractmethod
    def SandboxWait(
        self,
        request: modal_proto.api_pb2.SandboxWaitRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SandboxWaitResponse: ...
    @abc.abstractmethod
    def SecretDelete(
        self,
        request: modal_proto.api_pb2.SecretDeleteRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Secrets"""
    @abc.abstractmethod
    def SecretGetOrCreate(
        self,
        request: modal_proto.api_pb2.SecretGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SecretGetOrCreateResponse: ...
    @abc.abstractmethod
    def SecretList(
        self,
        request: modal_proto.api_pb2.SecretListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SecretListResponse: ...
    @abc.abstractmethod
    def SharedVolumeDelete(
        self,
        request: modal_proto.api_pb2.SharedVolumeDeleteRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """SharedVolumes"""
    @abc.abstractmethod
    def SharedVolumeGetFile(
        self,
        request: modal_proto.api_pb2.SharedVolumeGetFileRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SharedVolumeGetFileResponse: ...
    @abc.abstractmethod
    def SharedVolumeGetOrCreate(
        self,
        request: modal_proto.api_pb2.SharedVolumeGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SharedVolumeGetOrCreateResponse: ...
    @abc.abstractmethod
    def SharedVolumeHeartbeat(
        self,
        request: modal_proto.api_pb2.SharedVolumeHeartbeatRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def SharedVolumeList(
        self,
        request: modal_proto.api_pb2.SharedVolumeListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SharedVolumeListResponse: ...
    @abc.abstractmethod
    def SharedVolumeListFiles(
        self,
        request: modal_proto.api_pb2.SharedVolumeListFilesRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SharedVolumeListFilesResponse: ...
    @abc.abstractmethod
    def SharedVolumeListFilesStream(
        self,
        request: modal_proto.api_pb2.SharedVolumeListFilesRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.SharedVolumeListFilesResponse]: ...
    @abc.abstractmethod
    def SharedVolumePutFile(
        self,
        request: modal_proto.api_pb2.SharedVolumePutFileRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.SharedVolumePutFileResponse: ...
    @abc.abstractmethod
    def SharedVolumeRemoveFile(
        self,
        request: modal_proto.api_pb2.SharedVolumeRemoveFileRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def TaskClusterHello(
        self,
        request: modal_proto.api_pb2.TaskClusterHelloRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.TaskClusterHelloResponse:
        """Tasks"""
    @abc.abstractmethod
    def TaskCurrentInputs(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.TaskCurrentInputsResponse: ...
    @abc.abstractmethod
    def TaskList(
        self,
        request: modal_proto.api_pb2.TaskListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.TaskListResponse: ...
    @abc.abstractmethod
    def TaskResult(
        self,
        request: modal_proto.api_pb2.TaskResultRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def TokenFlowCreate(
        self,
        request: modal_proto.api_pb2.TokenFlowCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.TokenFlowCreateResponse:
        """Tokens (web auth flow)"""
    @abc.abstractmethod
    def TokenFlowWait(
        self,
        request: modal_proto.api_pb2.TokenFlowWaitRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.TokenFlowWaitResponse: ...
    @abc.abstractmethod
    def TunnelStart(
        self,
        request: modal_proto.api_pb2.TunnelStartRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.TunnelStartResponse:
        """Tunnels"""
    @abc.abstractmethod
    def TunnelStop(
        self,
        request: modal_proto.api_pb2.TunnelStopRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.TunnelStopResponse: ...
    @abc.abstractmethod
    def VolumeCommit(
        self,
        request: modal_proto.api_pb2.VolumeCommitRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.VolumeCommitResponse:
        """Volumes"""
    @abc.abstractmethod
    def VolumeCopyFiles(
        self,
        request: modal_proto.api_pb2.VolumeCopyFilesRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def VolumeDelete(
        self,
        request: modal_proto.api_pb2.VolumeDeleteRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def VolumeGetFile(
        self,
        request: modal_proto.api_pb2.VolumeGetFileRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.VolumeGetFileResponse: ...
    @abc.abstractmethod
    def VolumeGetOrCreate(
        self,
        request: modal_proto.api_pb2.VolumeGetOrCreateRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.VolumeGetOrCreateResponse: ...
    @abc.abstractmethod
    def VolumeHeartbeat(
        self,
        request: modal_proto.api_pb2.VolumeHeartbeatRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def VolumeList(
        self,
        request: modal_proto.api_pb2.VolumeListRequest,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.VolumeListResponse: ...
    @abc.abstractmethod
    def VolumeListFiles(
        self,
        request: modal_proto.api_pb2.VolumeListFilesRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[modal_proto.api_pb2.VolumeListFilesResponse]: ...
    @abc.abstractmethod
    def VolumePutFiles(
        self,
        request: modal_proto.api_pb2.VolumePutFilesRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def VolumeReload(
        self,
        request: modal_proto.api_pb2.VolumeReloadRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def VolumeRemoveFile(
        self,
        request: modal_proto.api_pb2.VolumeRemoveFileRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def VolumeRename(
        self,
        request: modal_proto.api_pb2.VolumeRenameRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...
    @abc.abstractmethod
    def WorkspaceNameLookup(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: grpc.ServicerContext,
    ) -> modal_proto.api_pb2.WorkspaceNameLookupResponse:
        """Workspaces"""

def add_ModalClientServicer_to_server(servicer: ModalClientServicer, server: grpc.Server) -> None: ...



# Copyright Modal Labs 2024
from collections.abc import Sequence
from typing import Optional, Union

from modal_proto import api_pb2


class SchedulerPlacement:
    """mdmd:hidden This is an experimental feature."""

    proto: api_pb2.SchedulerPlacement

    def __init__(
        self,
        region: Optional[Union[str, Sequence[str]]] = None,
        zone: Optional[str] = None,
        spot: Optional[bool] = None,
        instance_type: Optional[Union[str, Sequence[str]]] = None,
    ):
        """mdmd:hidden"""
        _lifecycle: Optional[str] = None
        if spot is not None:
            _lifecycle = "spot" if spot else "on-demand"

        regions = []
        if region:
            if isinstance(region, str):
                regions = [region]
            else:
                regions = list(region)

        instance_types = []
        if instance_type:
            if isinstance(instance_type, str):
                instance_types = [instance_type]
            else:
                instance_types = list(instance_type)

        self.proto = api_pb2.SchedulerPlacement(
            regions=regions,
            _zone=zone,
            _lifecycle=_lifecycle,
            _instance_types=instance_types,
        )



import asyncio
import asyncio.locks
import asyncio.queues
import collections.abc
import modal._runtime.user_code_imports
import modal.client
import modal_proto.api_pb2
import synchronicity.combined_types
import typing
import typing_extensions

class UserException(Exception): ...
class Sentinel: ...

class IOContext:
    input_ids: list[str]
    function_call_ids: list[str]
    finalized_function: modal._runtime.user_code_imports.FinalizedFunction
    _cancel_issued: bool
    _cancel_callback: typing.Optional[collections.abc.Callable[[], None]]

    def __init__(
        self,
        input_ids: list[str],
        function_call_ids: list[str],
        finalized_function: modal._runtime.user_code_imports.FinalizedFunction,
        function_inputs: list[modal_proto.api_pb2.FunctionInput],
        is_batched: bool,
        client: modal.client._Client,
    ): ...
    @classmethod
    async def create(
        cls,
        client: modal.client._Client,
        finalized_functions: dict[str, modal._runtime.user_code_imports.FinalizedFunction],
        inputs: list[tuple[str, str, modal_proto.api_pb2.FunctionInput]],
        is_batched: bool,
    ) -> IOContext: ...
    def set_cancel_callback(self, cb: collections.abc.Callable[[], None]): ...
    def cancel(self): ...
    def _args_and_kwargs(self) -> tuple[tuple[typing.Any, ...], dict[str, list[typing.Any]]]: ...
    def call_finalized_function(self) -> typing.Any: ...
    def validate_output_data(self, data: typing.Any) -> list[typing.Any]: ...

class InputSlots:
    active: int
    value: int
    waiter: typing.Optional[asyncio.Future]
    closed: bool

    def __init__(self, value: int) -> None: ...
    async def acquire(self) -> None: ...
    def _wake_waiter(self) -> None: ...
    def release(self) -> None: ...
    def set_value(self, value: int) -> None: ...
    async def close(self) -> None: ...

class _ContainerIOManager:
    task_id: str
    function_id: str
    app_id: str
    function_def: modal_proto.api_pb2.Function
    checkpoint_id: typing.Optional[str]
    calls_completed: int
    total_user_time: float
    current_input_id: typing.Optional[str]
    current_inputs: dict[str, IOContext]
    current_input_started_at: typing.Optional[float]
    _target_concurrency: int
    _max_concurrency: int
    _concurrency_loop: typing.Optional[asyncio.Task]
    _input_slots: InputSlots
    _environment_name: str
    _heartbeat_loop: typing.Optional[asyncio.Task]
    _heartbeat_condition: typing.Optional[asyncio.locks.Condition]
    _waiting_for_memory_snapshot: bool
    _is_interactivity_enabled: bool
    _fetching_inputs: bool
    _client: modal.client._Client
    _GENERATOR_STOP_SENTINEL: typing.ClassVar[Sentinel]
    _singleton: typing.ClassVar[typing.Optional[_ContainerIOManager]]

    def _init(self, container_args: modal_proto.api_pb2.ContainerArguments, client: modal.client._Client): ...
    @property
    def heartbeat_condition(self) -> asyncio.locks.Condition: ...
    @staticmethod
    def __new__(
        cls, container_args: modal_proto.api_pb2.ContainerArguments, client: modal.client._Client
    ) -> _ContainerIOManager: ...
    @classmethod
    def _reset_singleton(cls): ...
    async def hello(self): ...
    async def _run_heartbeat_loop(self): ...
    async def _heartbeat_handle_cancellations(self) -> bool: ...
    def heartbeats(self, wait_for_mem_snap: bool) -> typing.AsyncContextManager[None]: ...
    def stop_heartbeat(self): ...
    def dynamic_concurrency_manager(self) -> typing.AsyncContextManager[None]: ...
    async def _dynamic_concurrency_loop(self): ...
    async def get_serialized_function(
        self,
    ) -> tuple[typing.Optional[typing.Any], typing.Optional[collections.abc.Callable[..., typing.Any]]]: ...
    def serialize(self, obj: typing.Any) -> bytes: ...
    def deserialize(self, data: bytes) -> typing.Any: ...
    def serialize_data_format(self, obj: typing.Any, data_format: int) -> bytes: ...
    async def format_blob_data(self, data: bytes) -> dict[str, typing.Any]: ...
    def get_data_in(self, function_call_id: str) -> collections.abc.AsyncIterator[typing.Any]: ...
    async def put_data_out(
        self, function_call_id: str, start_index: int, data_format: int, serialized_messages: list[typing.Any]
    ) -> None: ...
    async def generator_output_task(
        self, function_call_id: str, data_format: int, message_rx: asyncio.queues.Queue
    ) -> None: ...
    async def _queue_create(self, size: int) -> asyncio.queues.Queue: ...
    async def _queue_put(self, queue: asyncio.queues.Queue, value: typing.Any) -> None: ...
    def get_average_call_time(self) -> float: ...
    def get_max_inputs_to_fetch(self): ...
    def _generate_inputs(
        self, batch_max_size: int, batch_wait_ms: int
    ) -> collections.abc.AsyncIterator[list[tuple[str, str, modal_proto.api_pb2.FunctionInput]]]: ...
    def run_inputs_outputs(
        self,
        finalized_functions: dict[str, modal._runtime.user_code_imports.FinalizedFunction],
        batch_max_size: int = 0,
        batch_wait_ms: int = 0,
    ) -> collections.abc.AsyncIterator[IOContext]: ...
    async def _push_outputs(
        self,
        io_context: IOContext,
        started_at: float,
        data_format: int,
        results: list[modal_proto.api_pb2.GenericResult],
    ) -> None: ...
    def serialize_exception(self, exc: BaseException) -> bytes: ...
    def serialize_traceback(self, exc: BaseException) -> tuple[typing.Optional[bytes], typing.Optional[bytes]]: ...
    def handle_user_exception(self) -> typing.AsyncContextManager[None]: ...
    def handle_input_exception(self, io_context: IOContext, started_at: float) -> typing.AsyncContextManager[None]: ...
    def exit_context(self, started_at, input_ids: list[str]): ...
    async def push_outputs(
        self, io_context: IOContext, started_at: float, data: typing.Any, data_format: int
    ) -> None: ...
    async def memory_restore(self) -> None: ...
    async def memory_snapshot(self) -> None: ...
    async def volume_commit(self, volume_ids: list[str]) -> None: ...
    async def interact(self, from_breakpoint: bool = False): ...
    @property
    def target_concurrency(self) -> int: ...
    @property
    def max_concurrency(self) -> int: ...
    @classmethod
    def get_input_concurrency(cls) -> int: ...
    @classmethod
    def set_input_concurrency(cls, concurrency: int): ...
    @classmethod
    def stop_fetching_inputs(cls): ...

SUPERSELF = typing.TypeVar("SUPERSELF", covariant=True)

class ContainerIOManager:
    task_id: str
    function_id: str
    app_id: str
    function_def: modal_proto.api_pb2.Function
    checkpoint_id: typing.Optional[str]
    calls_completed: int
    total_user_time: float
    current_input_id: typing.Optional[str]
    current_inputs: dict[str, IOContext]
    current_input_started_at: typing.Optional[float]
    _target_concurrency: int
    _max_concurrency: int
    _concurrency_loop: typing.Optional[asyncio.Task]
    _input_slots: InputSlots
    _environment_name: str
    _heartbeat_loop: typing.Optional[asyncio.Task]
    _heartbeat_condition: typing.Optional[asyncio.locks.Condition]
    _waiting_for_memory_snapshot: bool
    _is_interactivity_enabled: bool
    _fetching_inputs: bool
    _client: modal.client.Client
    _GENERATOR_STOP_SENTINEL: typing.ClassVar[Sentinel]
    _singleton: typing.ClassVar[typing.Optional[ContainerIOManager]]

    def __init__(self, /, *args, **kwargs): ...
    def _init(self, container_args: modal_proto.api_pb2.ContainerArguments, client: modal.client.Client): ...
    @property
    def heartbeat_condition(self) -> asyncio.locks.Condition: ...
    @classmethod
    def _reset_singleton(cls): ...

    class __hello_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self): ...
        async def aio(self): ...

    hello: __hello_spec[typing_extensions.Self]

    class ___run_heartbeat_loop_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self): ...
        async def aio(self): ...

    _run_heartbeat_loop: ___run_heartbeat_loop_spec[typing_extensions.Self]

    class ___heartbeat_handle_cancellations_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self) -> bool: ...
        async def aio(self) -> bool: ...

    _heartbeat_handle_cancellations: ___heartbeat_handle_cancellations_spec[typing_extensions.Self]

    class __heartbeats_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(
            self, wait_for_mem_snap: bool
        ) -> synchronicity.combined_types.AsyncAndBlockingContextManager[None]: ...
        def aio(self, wait_for_mem_snap: bool) -> typing.AsyncContextManager[None]: ...

    heartbeats: __heartbeats_spec[typing_extensions.Self]

    def stop_heartbeat(self): ...

    class __dynamic_concurrency_manager_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self) -> synchronicity.combined_types.AsyncAndBlockingContextManager[None]: ...
        def aio(self) -> typing.AsyncContextManager[None]: ...

    dynamic_concurrency_manager: __dynamic_concurrency_manager_spec[typing_extensions.Self]

    class ___dynamic_concurrency_loop_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self): ...
        async def aio(self): ...

    _dynamic_concurrency_loop: ___dynamic_concurrency_loop_spec[typing_extensions.Self]

    class __get_serialized_function_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(
            self,
        ) -> tuple[typing.Optional[typing.Any], typing.Optional[collections.abc.Callable[..., typing.Any]]]: ...
        async def aio(
            self,
        ) -> tuple[typing.Optional[typing.Any], typing.Optional[collections.abc.Callable[..., typing.Any]]]: ...

    get_serialized_function: __get_serialized_function_spec[typing_extensions.Self]

    def serialize(self, obj: typing.Any) -> bytes: ...
    def deserialize(self, data: bytes) -> typing.Any: ...
    def serialize_data_format(self, obj: typing.Any, data_format: int) -> bytes: ...

    class __format_blob_data_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, data: bytes) -> dict[str, typing.Any]: ...
        async def aio(self, data: bytes) -> dict[str, typing.Any]: ...

    format_blob_data: __format_blob_data_spec[typing_extensions.Self]

    class __get_data_in_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, function_call_id: str) -> typing.Iterator[typing.Any]: ...
        def aio(self, function_call_id: str) -> collections.abc.AsyncIterator[typing.Any]: ...

    get_data_in: __get_data_in_spec[typing_extensions.Self]

    class __put_data_out_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(
            self, function_call_id: str, start_index: int, data_format: int, serialized_messages: list[typing.Any]
        ) -> None: ...
        async def aio(
            self, function_call_id: str, start_index: int, data_format: int, serialized_messages: list[typing.Any]
        ) -> None: ...

    put_data_out: __put_data_out_spec[typing_extensions.Self]

    class __generator_output_task_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, function_call_id: str, data_format: int, message_rx: asyncio.queues.Queue) -> None: ...
        async def aio(self, function_call_id: str, data_format: int, message_rx: asyncio.queues.Queue) -> None: ...

    generator_output_task: __generator_output_task_spec[typing_extensions.Self]

    class ___queue_create_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, size: int) -> asyncio.queues.Queue: ...
        async def aio(self, size: int) -> asyncio.queues.Queue: ...

    _queue_create: ___queue_create_spec[typing_extensions.Self]

    class ___queue_put_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, queue: asyncio.queues.Queue, value: typing.Any) -> None: ...
        async def aio(self, queue: asyncio.queues.Queue, value: typing.Any) -> None: ...

    _queue_put: ___queue_put_spec[typing_extensions.Self]

    def get_average_call_time(self) -> float: ...
    def get_max_inputs_to_fetch(self): ...

    class ___generate_inputs_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(
            self, batch_max_size: int, batch_wait_ms: int
        ) -> typing.Iterator[list[tuple[str, str, modal_proto.api_pb2.FunctionInput]]]: ...
        def aio(
            self, batch_max_size: int, batch_wait_ms: int
        ) -> collections.abc.AsyncIterator[list[tuple[str, str, modal_proto.api_pb2.FunctionInput]]]: ...

    _generate_inputs: ___generate_inputs_spec[typing_extensions.Self]

    class __run_inputs_outputs_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(
            self,
            finalized_functions: dict[str, modal._runtime.user_code_imports.FinalizedFunction],
            batch_max_size: int = 0,
            batch_wait_ms: int = 0,
        ) -> typing.Iterator[IOContext]: ...
        def aio(
            self,
            finalized_functions: dict[str, modal._runtime.user_code_imports.FinalizedFunction],
            batch_max_size: int = 0,
            batch_wait_ms: int = 0,
        ) -> collections.abc.AsyncIterator[IOContext]: ...

    run_inputs_outputs: __run_inputs_outputs_spec[typing_extensions.Self]

    class ___push_outputs_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(
            self,
            io_context: IOContext,
            started_at: float,
            data_format: int,
            results: list[modal_proto.api_pb2.GenericResult],
        ) -> None: ...
        async def aio(
            self,
            io_context: IOContext,
            started_at: float,
            data_format: int,
            results: list[modal_proto.api_pb2.GenericResult],
        ) -> None: ...

    _push_outputs: ___push_outputs_spec[typing_extensions.Self]

    def serialize_exception(self, exc: BaseException) -> bytes: ...
    def serialize_traceback(self, exc: BaseException) -> tuple[typing.Optional[bytes], typing.Optional[bytes]]: ...

    class __handle_user_exception_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self) -> synchronicity.combined_types.AsyncAndBlockingContextManager[None]: ...
        def aio(self) -> typing.AsyncContextManager[None]: ...

    handle_user_exception: __handle_user_exception_spec[typing_extensions.Self]

    class __handle_input_exception_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(
            self, io_context: IOContext, started_at: float
        ) -> synchronicity.combined_types.AsyncAndBlockingContextManager[None]: ...
        def aio(self, io_context: IOContext, started_at: float) -> typing.AsyncContextManager[None]: ...

    handle_input_exception: __handle_input_exception_spec[typing_extensions.Self]

    def exit_context(self, started_at, input_ids: list[str]): ...

    class __push_outputs_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, io_context: IOContext, started_at: float, data: typing.Any, data_format: int) -> None: ...
        async def aio(self, io_context: IOContext, started_at: float, data: typing.Any, data_format: int) -> None: ...

    push_outputs: __push_outputs_spec[typing_extensions.Self]

    class __memory_restore_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self) -> None: ...
        async def aio(self) -> None: ...

    memory_restore: __memory_restore_spec[typing_extensions.Self]

    class __memory_snapshot_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self) -> None: ...
        async def aio(self) -> None: ...

    memory_snapshot: __memory_snapshot_spec[typing_extensions.Self]

    class __volume_commit_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, volume_ids: list[str]) -> None: ...
        async def aio(self, volume_ids: list[str]) -> None: ...

    volume_commit: __volume_commit_spec[typing_extensions.Self]

    class __interact_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, from_breakpoint: bool = False): ...
        async def aio(self, from_breakpoint: bool = False): ...

    interact: __interact_spec[typing_extensions.Self]

    @property
    def target_concurrency(self) -> int: ...
    @property
    def max_concurrency(self) -> int: ...
    @classmethod
    def get_input_concurrency(cls) -> int: ...
    @classmethod
    def set_input_concurrency(cls, concurrency: int): ...
    @classmethod
    def stop_fetching_inputs(cls): ...

def check_fastapi_pydantic_compatibility(exc: ImportError) -> None: ...

MAX_OUTPUT_BATCH_SIZE: int

RTT_S: float



# Copyright Modal Labs 2022
import typing
import uuid
from collections.abc import Awaitable, Hashable, Sequence
from functools import wraps
from typing import Callable, ClassVar, Optional

from google.protobuf.message import Message
from typing_extensions import Self

from ._resolver import Resolver
from ._utils.async_utils import aclosing
from ._utils.deprecation import deprecation_warning
from .client import _Client
from .config import config, logger
from .exception import ExecutionError, InvalidError

EPHEMERAL_OBJECT_HEARTBEAT_SLEEP: int = 300


def _get_environment_name(environment_name: Optional[str] = None, resolver: Optional[Resolver] = None) -> Optional[str]:
    if environment_name:
        return environment_name
    elif resolver and resolver.environment_name:
        return resolver.environment_name
    else:
        return config.get("environment")


class _Object:
    _type_prefix: ClassVar[Optional[str]] = None
    _prefix_to_type: ClassVar[dict[str, type]] = {}

    # For constructors
    _load: Optional[Callable[[Self, Resolver, Optional[str]], Awaitable[None]]]
    _preload: Optional[Callable[[Self, Resolver, Optional[str]], Awaitable[None]]]
    _rep: str
    _is_another_app: bool
    _hydrate_lazily: bool
    _deps: Optional[Callable[..., Sequence["_Object"]]]
    _deduplication_key: Optional[Callable[[], Awaitable[Hashable]]] = None

    # For hydrated objects
    _object_id: Optional[str]
    _client: Optional[_Client]
    _is_hydrated: bool
    _is_rehydrated: bool

    @classmethod
    def __init_subclass__(cls, type_prefix: Optional[str] = None):
        super().__init_subclass__()
        if type_prefix is not None:
            cls._type_prefix = type_prefix
            cls._prefix_to_type[type_prefix] = cls

    def __init__(self, *args, **kwargs):
        raise InvalidError(f"Class {type(self).__name__} has no constructor. Use class constructor methods instead.")

    def _init(
        self,
        rep: str,
        load: Optional[Callable[[Self, Resolver, Optional[str]], Awaitable[None]]] = None,
        is_another_app: bool = False,
        preload: Optional[Callable[[Self, Resolver, Optional[str]], Awaitable[None]]] = None,
        hydrate_lazily: bool = False,
        deps: Optional[Callable[..., Sequence["_Object"]]] = None,
        deduplication_key: Optional[Callable[[], Awaitable[Hashable]]] = None,
    ):
        self._local_uuid = str(uuid.uuid4())
        self._load = load
        self._preload = preload
        self._rep = rep
        self._is_another_app = is_another_app
        self._hydrate_lazily = hydrate_lazily
        self._deps = deps
        self._deduplication_key = deduplication_key

        self._object_id = None
        self._client = None
        self._is_hydrated = False
        self._is_rehydrated = False

        self._initialize_from_empty()

    def _unhydrate(self):
        self._object_id = None
        self._client = None
        self._is_hydrated = False

    def _initialize_from_empty(self):
        # default implementation, can be overriden in subclasses
        pass

    def _initialize_from_other(self, other):
        # default implementation, can be overriden in subclasses
        self._object_id = other._object_id
        self._is_hydrated = other._is_hydrated
        self._client = other._client

    def _hydrate(self, object_id: str, client: _Client, metadata: Optional[Message]):
        assert isinstance(object_id, str) and self._type_prefix is not None
        if not object_id.startswith(self._type_prefix):
            raise ExecutionError(
                f"Can not hydrate {type(self)}:"
                f" it has type prefix {self._type_prefix}"
                f" but the object_id starts with {object_id[:3]}"
            )
        self._object_id = object_id
        self._client = client
        self._hydrate_metadata(metadata)
        self._is_hydrated = True

    def _hydrate_metadata(self, metadata: Optional[Message]):
        # override this is subclasses that need additional data (other than an object_id) for a functioning Handle
        pass

    def _get_metadata(self) -> Optional[Message]:
        # return the necessary metadata from this handle to be able to re-hydrate in another context if one is needed
        # used to provide a handle's handle_metadata for serializing/pickling a live handle
        # the object_id is already provided by other means
        return None

    def _validate_is_hydrated(self):
        if not self._is_hydrated:
            object_type = self.__class__.__name__.strip("_")
            if hasattr(self, "_app") and getattr(self._app, "_running_app", "") is None:  # type: ignore
                # The most common cause of this error: e.g., user called a Function without using App.run()
                reason = ", because the App it is defined on is not running"
            else:
                # Technically possible, but with an ambiguous cause.
                reason = ""
            raise ExecutionError(
                f"{object_type} has not been hydrated with the metadata it needs to run on Modal{reason}."
            )

    def clone(self) -> Self:
        """mdmd:hidden Clone a given hydrated object."""

        # Object to clone must already be hydrated, otherwise from_loader is more suitable.
        self._validate_is_hydrated()
        obj = type(self).__new__(type(self))
        obj._initialize_from_other(self)
        return obj

    @classmethod
    def _from_loader(
        cls,
        load: Callable[[Self, Resolver, Optional[str]], Awaitable[None]],
        rep: str,
        is_another_app: bool = False,
        preload: Optional[Callable[[Self, Resolver, Optional[str]], Awaitable[None]]] = None,
        hydrate_lazily: bool = False,
        deps: Optional[Callable[..., Sequence["_Object"]]] = None,
        deduplication_key: Optional[Callable[[], Awaitable[Hashable]]] = None,
    ):
        # TODO(erikbern): flip the order of the two first arguments
        obj = _Object.__new__(cls)
        obj._init(rep, load, is_another_app, preload, hydrate_lazily, deps, deduplication_key)
        return obj

    @staticmethod
    def _get_type_from_id(object_id: str) -> type["_Object"]:
        parts = object_id.split("-")
        if len(parts) != 2:
            raise InvalidError(f"Object id {object_id} has no dash in it")
        prefix = parts[0]
        if prefix not in _Object._prefix_to_type:
            raise InvalidError(f"Object prefix {prefix} does not correspond to a type")
        return _Object._prefix_to_type[prefix]

    @classmethod
    def _is_id_type(cls, object_id) -> bool:
        return cls._get_type_from_id(object_id) == cls

    @classmethod
    def _new_hydrated(
        cls, object_id: str, client: _Client, handle_metadata: Optional[Message], is_another_app: bool = False
    ) -> Self:
        obj_cls: type[Self]
        if cls._type_prefix is not None:
            # This is called directly on a subclass, e.g. Secret.from_id
            # validate the id matching the expected id type of the Object subclass
            if not object_id.startswith(cls._type_prefix + "-"):
                raise InvalidError(f"Object {object_id} does not start with {cls._type_prefix}")

            obj_cls = cls
        else:
            # this means the method is used directly on _Object
            # typically during deserialization of objects
            obj_cls = typing.cast(type[Self], cls._get_type_from_id(object_id))

        # Instantiate provider
        obj = _Object.__new__(obj_cls)
        rep = f"Object({object_id})"  # TODO(erikbern): dumb
        obj._init(rep, is_another_app=is_another_app)
        obj._hydrate(object_id, client, handle_metadata)

        return obj

    def _hydrate_from_other(self, other: Self):
        self._hydrate(other.object_id, other.client, other._get_metadata())

    def __repr__(self):
        return self._rep

    @property
    def local_uuid(self):
        """mdmd:hidden"""
        return self._local_uuid

    @property
    def object_id(self) -> str:
        """mdmd:hidden"""
        if self._object_id is None:
            raise AttributeError(f"Attempting to get object_id of unhydrated {self}")
        return self._object_id

    @property
    def client(self) -> _Client:
        """mdmd:hidden"""
        if self._client is None:
            raise AttributeError(f"Attempting to get client of unhydrated {self}")
        return self._client

    @property
    def is_hydrated(self) -> bool:
        """mdmd:hidden"""
        return self._is_hydrated

    @property
    def deps(self) -> Callable[..., Sequence["_Object"]]:
        """mdmd:hidden"""

        def default_deps(*args, **kwargs) -> Sequence["_Object"]:
            return []

        return self._deps if self._deps is not None else default_deps

    async def resolve(self, client: Optional[_Client] = None):
        """mdmd:hidden"""
        obj = self.__class__.__name__.strip("_")
        deprecation_warning(
            (2025, 1, 16),
            f"The `{obj}.resolve` method is deprecated and will be removed in a future release."
            f" Please use `{obj}.hydrate()` or `await {obj}.hydrate.aio()` instead."
            "\n\nNote that it is rarely necessary to explicitly hydrate objects, as most methods"
            " will lazily hydrate when needed.",
            show_source=False,  # synchronicity interferes with attributing source correctly
        )
        await self.hydrate(client)

    async def hydrate(self, client: Optional[_Client] = None) -> Self:
        """Synchronize the local object with its identity on the Modal server.

        It is rarely necessary to call this method explicitly, as most operations
        will lazily hydrate when needed. The main use case is when you need to
        access object metadata, such as its ID.

        *Added in v0.72.39*: This method replaces the deprecated `.resolve()` method.
        """
        if self._is_hydrated:
            if self.client._snapshotted and not self._is_rehydrated:
                # memory snapshots capture references which must be rehydrated
                # on restore to handle staleness.
                logger.debug(f"rehydrating {self} after snapshot")
                self._is_hydrated = False  # un-hydrate and re-resolve
                c = client if client is not None else await _Client.from_env()
                resolver = Resolver(c)
                await resolver.load(typing.cast(_Object, self))
                self._is_rehydrated = True
                logger.debug(f"rehydrated {self} with client {id(c)}")
        elif not self._hydrate_lazily:
            # TODO(michael) can remove _hydrate lazily? I think all objects support it now?
            self._validate_is_hydrated()
        else:
            c = client if client is not None else await _Client.from_env()
            resolver = Resolver(c)
            await resolver.load(self)
        return self


def live_method(method):
    @wraps(method)
    async def wrapped(self, *args, **kwargs):
        await self.hydrate()
        return await method(self, *args, **kwargs)

    return wrapped


def live_method_gen(method):
    @wraps(method)
    async def wrapped(self, *args, **kwargs):
        await self.hydrate()
        async with aclosing(method(self, *args, **kwargs)) as stream:
            async for item in stream:
                yield item

    return wrapped



# Copyright Modal Labs 2022
"""Helper functions related to operating on exceptions, warnings, and traceback objects.

Functions related to *displaying* tracebacks should go in `modal/cli/_traceback.py`
so that Rich is not a dependency of the container Client.
"""

import re
import sys
import traceback
import warnings
from types import TracebackType
from typing import Any, Iterable, Optional

from modal_proto import api_pb2

from ._vendor.tblib import Traceback as TBLibTraceback
from .exception import ServerWarning

TBDictType = dict[str, Any]
LineCacheType = dict[tuple[str, str], str]


def extract_traceback(exc: BaseException, task_id: str) -> tuple[TBDictType, LineCacheType]:
    """Given an exception, extract a serializable traceback (with task ID markers included),
    and a line cache that maps (filename, lineno) to line contents. The latter is used to show
    a helpful traceback to the user, even if they don't have packages installed locally that
    are referenced in the traceback."""

    tb = TBLibTraceback(exc.__traceback__)
    # Prefix traceback file paths with <task_id>. This lets us attribute which parts of
    # the traceback came from specific remote containers, while still fitting in the TracebackType
    # spec. Real paths can never start with <; we can use this to extract task_ids from filenames
    # at the client.
    cur = tb
    while cur is not None:
        file = cur.tb_frame.f_code.co_filename

        # Paths starting with < indicate that they're from a traceback from a remote
        # container. This means we've reached the end of the local traceback.
        if file.startswith("<"):
            break
        # We rely on this specific filename format when inferring where the exception was raised
        # in various other exception-related code
        cur.tb_frame.f_code.co_filename = f"<{task_id}>:{file}"
        cur = cur.tb_next

    tb_dict = tb.to_dict()

    line_cache = getattr(exc, "__line_cache__", {})

    for frame in traceback.extract_tb(exc.__traceback__):
        line_cache[(frame.filename, frame.lineno)] = frame.line

    return tb_dict, line_cache


def append_modal_tb(exc: BaseException, tb_dict: TBDictType, line_cache: LineCacheType) -> None:
    tb = TBLibTraceback.from_dict(tb_dict).as_traceback()

    # Filter out the prefix corresponding to internal Modal frames, and then make
    # the remote traceback from a Modal function the starting point of the current
    # exception's traceback.

    while tb is not None:
        if "/pkg/modal/" not in tb.tb_frame.f_code.co_filename:
            break
        tb = tb.tb_next

    exc.__traceback__ = tb

    setattr(exc, "__line_cache__", line_cache)


def reduce_traceback_to_user_code(tb: Optional[TracebackType], user_source: str) -> TracebackType:
    """Return a traceback that does not contain modal entrypoint or synchronicity frames."""

    # Step forward all the way through the traceback and drop any "Modal support" frames
    def skip_frame(filename: str) -> bool:
        return "/site-packages/synchronicity/" in filename or "modal/_utils/deprecation" in filename

    tb_root = tb
    while tb is not None:
        while tb.tb_next is not None:
            if skip_frame(tb.tb_next.tb_frame.f_code.co_filename):
                tb.tb_next = tb.tb_next.tb_next
            else:
                break
        tb = tb.tb_next
    tb = tb_root

    # Now step forward again until we get to first frame of user code
    if user_source.endswith(".py"):
        while tb is not None and tb.tb_frame.f_code.co_filename != user_source:
            tb = tb.tb_next
    else:
        while tb is not None and tb.tb_frame.f_code.co_name != "<module>":
            tb = tb.tb_next
    if tb is None:
        # In case we didn't find a frame that matched the user source, revert to the original root
        tb = tb_root

    return tb


def traceback_contains_remote_call(tb: Optional[TracebackType]) -> bool:
    """Inspect the traceback stack to determine whether an error was raised locally or remotely."""
    while tb is not None:
        if re.match(r"^<ta-[0-9A-Z]{26}>:", tb.tb_frame.f_code.co_filename):
            return True
        tb = tb.tb_next
    return False


def print_exception(exc: Optional[type[BaseException]], value: Optional[BaseException], tb: Optional[TracebackType]):
    """Add backwards compatibility for printing exceptions with "notes" for Python<3.11."""
    traceback.print_exception(exc, value, tb)
    if sys.version_info < (3, 11) and value is not None:
        notes = getattr(value, "__notes__", [])
        print(*notes, sep="\n", file=sys.stderr)


def print_server_warnings(server_warnings: Iterable[api_pb2.Warning]):
    """Issue a warning originating from the server with empty metadata about local origin.

    When using the Modal CLI, these warnings should get caught and coerced into Rich panels.
    """
    for warning in server_warnings:
        warnings.warn_explicit(warning.message, ServerWarning, "<modal-server>", 0)



import modal._clustered_functions
import modal.client
import typing
import typing_extensions

def stop_fetching_inputs(): ...
def get_local_input_concurrency(): ...
def set_local_input_concurrency(concurrency: int): ...
def clustered(size: int, broadcast: bool = True): ...
def get_cluster_info() -> modal._clustered_functions.ClusterInfo: ...

class AppInfo:
    app_id: str
    name: str
    containers: int

    def __init__(self, app_id: str, name: str, containers: int) -> None: ...
    def __repr__(self): ...
    def __eq__(self, other): ...

class __list_deployed_apps_spec(typing_extensions.Protocol):
    def __call__(
        self, environment_name: str = "", client: typing.Optional[modal.client.Client] = None
    ) -> list[AppInfo]: ...
    async def aio(
        self, environment_name: str = "", client: typing.Optional[modal.client.Client] = None
    ) -> list[AppInfo]: ...

list_deployed_apps: __list_deployed_apps_spec



# Generated by the Modal Protocol Buffers compiler. DO NOT EDIT!
# source: modal_proto/api.proto
# plugin: __main__

import modal._utils.grpc_utils
import modal_proto.api_grpc
import typing
if typing.TYPE_CHECKING:
    import modal.client


class ModalClientModal:

    def __init__(self, grpclib_stub: modal_proto.api_grpc.ModalClientStub, client: "modal.client._Client") -> None:
        self.AppClientDisconnect = modal.client.UnaryUnaryWrapper(grpclib_stub.AppClientDisconnect, client)
        self.AppCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.AppCreate, client)
        self.AppDeploy = modal.client.UnaryUnaryWrapper(grpclib_stub.AppDeploy, client)
        self.AppDeploymentHistory = modal.client.UnaryUnaryWrapper(grpclib_stub.AppDeploymentHistory, client)
        self.AppGetByDeploymentName = modal.client.UnaryUnaryWrapper(grpclib_stub.AppGetByDeploymentName, client)
        self.AppGetLayout = modal.client.UnaryUnaryWrapper(grpclib_stub.AppGetLayout, client)
        self.AppGetLogs = modal.client.UnaryStreamWrapper(grpclib_stub.AppGetLogs, client)
        self.AppGetObjects = modal.client.UnaryUnaryWrapper(grpclib_stub.AppGetObjects, client)
        self.AppGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.AppGetOrCreate, client)
        self.AppHeartbeat = modal.client.UnaryUnaryWrapper(grpclib_stub.AppHeartbeat, client)
        self.AppList = modal.client.UnaryUnaryWrapper(grpclib_stub.AppList, client)
        self.AppLookup = modal.client.UnaryUnaryWrapper(grpclib_stub.AppLookup, client)
        self.AppPublish = modal.client.UnaryUnaryWrapper(grpclib_stub.AppPublish, client)
        self.AppRollback = modal.client.UnaryUnaryWrapper(grpclib_stub.AppRollback, client)
        self.AppSetObjects = modal.client.UnaryUnaryWrapper(grpclib_stub.AppSetObjects, client)
        self.AppStop = modal.client.UnaryUnaryWrapper(grpclib_stub.AppStop, client)
        self.BlobCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.BlobCreate, client)
        self.BlobGet = modal.client.UnaryUnaryWrapper(grpclib_stub.BlobGet, client)
        self.ClassCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.ClassCreate, client)
        self.ClassGet = modal.client.UnaryUnaryWrapper(grpclib_stub.ClassGet, client)
        self.ClientHello = modal.client.UnaryUnaryWrapper(grpclib_stub.ClientHello, client)
        self.ContainerCheckpoint = modal.client.UnaryUnaryWrapper(grpclib_stub.ContainerCheckpoint, client)
        self.ContainerExec = modal.client.UnaryUnaryWrapper(grpclib_stub.ContainerExec, client)
        self.ContainerExecGetOutput = modal.client.UnaryStreamWrapper(grpclib_stub.ContainerExecGetOutput, client)
        self.ContainerExecPutInput = modal.client.UnaryUnaryWrapper(grpclib_stub.ContainerExecPutInput, client)
        self.ContainerExecWait = modal.client.UnaryUnaryWrapper(grpclib_stub.ContainerExecWait, client)
        self.ContainerFilesystemExec = modal.client.UnaryUnaryWrapper(grpclib_stub.ContainerFilesystemExec, client)
        self.ContainerFilesystemExecGetOutput = modal.client.UnaryStreamWrapper(grpclib_stub.ContainerFilesystemExecGetOutput, client)
        self.ContainerHeartbeat = modal.client.UnaryUnaryWrapper(grpclib_stub.ContainerHeartbeat, client)
        self.ContainerHello = modal.client.UnaryUnaryWrapper(grpclib_stub.ContainerHello, client)
        self.ContainerLog = modal.client.UnaryUnaryWrapper(grpclib_stub.ContainerLog, client)
        self.ContainerStop = modal.client.UnaryUnaryWrapper(grpclib_stub.ContainerStop, client)
        self.DictClear = modal.client.UnaryUnaryWrapper(grpclib_stub.DictClear, client)
        self.DictContains = modal.client.UnaryUnaryWrapper(grpclib_stub.DictContains, client)
        self.DictContents = modal.client.UnaryStreamWrapper(grpclib_stub.DictContents, client)
        self.DictDelete = modal.client.UnaryUnaryWrapper(grpclib_stub.DictDelete, client)
        self.DictGet = modal.client.UnaryUnaryWrapper(grpclib_stub.DictGet, client)
        self.DictGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.DictGetOrCreate, client)
        self.DictHeartbeat = modal.client.UnaryUnaryWrapper(grpclib_stub.DictHeartbeat, client)
        self.DictLen = modal.client.UnaryUnaryWrapper(grpclib_stub.DictLen, client)
        self.DictList = modal.client.UnaryUnaryWrapper(grpclib_stub.DictList, client)
        self.DictPop = modal.client.UnaryUnaryWrapper(grpclib_stub.DictPop, client)
        self.DictUpdate = modal.client.UnaryUnaryWrapper(grpclib_stub.DictUpdate, client)
        self.DomainCertificateVerify = modal.client.UnaryUnaryWrapper(grpclib_stub.DomainCertificateVerify, client)
        self.DomainCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.DomainCreate, client)
        self.DomainList = modal.client.UnaryUnaryWrapper(grpclib_stub.DomainList, client)
        self.EnvironmentCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.EnvironmentCreate, client)
        self.EnvironmentDelete = modal.client.UnaryUnaryWrapper(grpclib_stub.EnvironmentDelete, client)
        self.EnvironmentGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.EnvironmentGetOrCreate, client)
        self.EnvironmentList = modal.client.UnaryUnaryWrapper(grpclib_stub.EnvironmentList, client)
        self.EnvironmentUpdate = modal.client.UnaryUnaryWrapper(grpclib_stub.EnvironmentUpdate, client)
        self.FunctionAsyncInvoke = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionAsyncInvoke, client)
        self.FunctionBindParams = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionBindParams, client)
        self.FunctionCallCancel = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionCallCancel, client)
        self.FunctionCallGetDataIn = modal.client.UnaryStreamWrapper(grpclib_stub.FunctionCallGetDataIn, client)
        self.FunctionCallGetDataOut = modal.client.UnaryStreamWrapper(grpclib_stub.FunctionCallGetDataOut, client)
        self.FunctionCallList = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionCallList, client)
        self.FunctionCallPutDataOut = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionCallPutDataOut, client)
        self.FunctionCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionCreate, client)
        self.FunctionGet = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionGet, client)
        self.FunctionGetCallGraph = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionGetCallGraph, client)
        self.FunctionGetCurrentStats = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionGetCurrentStats, client)
        self.FunctionGetDynamicConcurrency = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionGetDynamicConcurrency, client)
        self.FunctionGetInputs = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionGetInputs, client)
        self.FunctionGetOutputs = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionGetOutputs, client)
        self.FunctionGetSerialized = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionGetSerialized, client)
        self.FunctionMap = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionMap, client)
        self.FunctionPrecreate = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionPrecreate, client)
        self.FunctionPutInputs = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionPutInputs, client)
        self.FunctionPutOutputs = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionPutOutputs, client)
        self.FunctionRetryInputs = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionRetryInputs, client)
        self.FunctionStartPtyShell = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionStartPtyShell, client)
        self.FunctionUpdateSchedulingParams = modal.client.UnaryUnaryWrapper(grpclib_stub.FunctionUpdateSchedulingParams, client)
        self.ImageFromId = modal.client.UnaryUnaryWrapper(grpclib_stub.ImageFromId, client)
        self.ImageGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.ImageGetOrCreate, client)
        self.ImageJoinStreaming = modal.client.UnaryStreamWrapper(grpclib_stub.ImageJoinStreaming, client)
        self.MountGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.MountGetOrCreate, client)
        self.MountPutFile = modal.client.UnaryUnaryWrapper(grpclib_stub.MountPutFile, client)
        self.NotebookKernelPublishResults = modal.client.UnaryUnaryWrapper(grpclib_stub.NotebookKernelPublishResults, client)
        self.ProxyAddIp = modal.client.UnaryUnaryWrapper(grpclib_stub.ProxyAddIp, client)
        self.ProxyCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.ProxyCreate, client)
        self.ProxyDelete = modal.client.UnaryUnaryWrapper(grpclib_stub.ProxyDelete, client)
        self.ProxyGet = modal.client.UnaryUnaryWrapper(grpclib_stub.ProxyGet, client)
        self.ProxyGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.ProxyGetOrCreate, client)
        self.ProxyList = modal.client.UnaryUnaryWrapper(grpclib_stub.ProxyList, client)
        self.ProxyRemoveIp = modal.client.UnaryUnaryWrapper(grpclib_stub.ProxyRemoveIp, client)
        self.QueueClear = modal.client.UnaryUnaryWrapper(grpclib_stub.QueueClear, client)
        self.QueueDelete = modal.client.UnaryUnaryWrapper(grpclib_stub.QueueDelete, client)
        self.QueueGet = modal.client.UnaryUnaryWrapper(grpclib_stub.QueueGet, client)
        self.QueueGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.QueueGetOrCreate, client)
        self.QueueHeartbeat = modal.client.UnaryUnaryWrapper(grpclib_stub.QueueHeartbeat, client)
        self.QueueLen = modal.client.UnaryUnaryWrapper(grpclib_stub.QueueLen, client)
        self.QueueList = modal.client.UnaryUnaryWrapper(grpclib_stub.QueueList, client)
        self.QueueNextItems = modal.client.UnaryUnaryWrapper(grpclib_stub.QueueNextItems, client)
        self.QueuePut = modal.client.UnaryUnaryWrapper(grpclib_stub.QueuePut, client)
        self.SandboxCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxCreate, client)
        self.SandboxGetLogs = modal.client.UnaryStreamWrapper(grpclib_stub.SandboxGetLogs, client)
        self.SandboxGetTaskId = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxGetTaskId, client)
        self.SandboxGetTunnels = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxGetTunnels, client)
        self.SandboxList = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxList, client)
        self.SandboxRestore = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxRestore, client)
        self.SandboxSnapshot = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxSnapshot, client)
        self.SandboxSnapshotFs = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxSnapshotFs, client)
        self.SandboxSnapshotGet = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxSnapshotGet, client)
        self.SandboxSnapshotWait = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxSnapshotWait, client)
        self.SandboxStdinWrite = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxStdinWrite, client)
        self.SandboxTagsSet = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxTagsSet, client)
        self.SandboxTerminate = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxTerminate, client)
        self.SandboxWait = modal.client.UnaryUnaryWrapper(grpclib_stub.SandboxWait, client)
        self.SecretDelete = modal.client.UnaryUnaryWrapper(grpclib_stub.SecretDelete, client)
        self.SecretGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.SecretGetOrCreate, client)
        self.SecretList = modal.client.UnaryUnaryWrapper(grpclib_stub.SecretList, client)
        self.SharedVolumeDelete = modal.client.UnaryUnaryWrapper(grpclib_stub.SharedVolumeDelete, client)
        self.SharedVolumeGetFile = modal.client.UnaryUnaryWrapper(grpclib_stub.SharedVolumeGetFile, client)
        self.SharedVolumeGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.SharedVolumeGetOrCreate, client)
        self.SharedVolumeHeartbeat = modal.client.UnaryUnaryWrapper(grpclib_stub.SharedVolumeHeartbeat, client)
        self.SharedVolumeList = modal.client.UnaryUnaryWrapper(grpclib_stub.SharedVolumeList, client)
        self.SharedVolumeListFiles = modal.client.UnaryUnaryWrapper(grpclib_stub.SharedVolumeListFiles, client)
        self.SharedVolumeListFilesStream = modal.client.UnaryStreamWrapper(grpclib_stub.SharedVolumeListFilesStream, client)
        self.SharedVolumePutFile = modal.client.UnaryUnaryWrapper(grpclib_stub.SharedVolumePutFile, client)
        self.SharedVolumeRemoveFile = modal.client.UnaryUnaryWrapper(grpclib_stub.SharedVolumeRemoveFile, client)
        self.TaskClusterHello = modal.client.UnaryUnaryWrapper(grpclib_stub.TaskClusterHello, client)
        self.TaskCurrentInputs = modal.client.UnaryUnaryWrapper(grpclib_stub.TaskCurrentInputs, client)
        self.TaskList = modal.client.UnaryUnaryWrapper(grpclib_stub.TaskList, client)
        self.TaskResult = modal.client.UnaryUnaryWrapper(grpclib_stub.TaskResult, client)
        self.TokenFlowCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.TokenFlowCreate, client)
        self.TokenFlowWait = modal.client.UnaryUnaryWrapper(grpclib_stub.TokenFlowWait, client)
        self.TunnelStart = modal.client.UnaryUnaryWrapper(grpclib_stub.TunnelStart, client)
        self.TunnelStop = modal.client.UnaryUnaryWrapper(grpclib_stub.TunnelStop, client)
        self.VolumeCommit = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeCommit, client)
        self.VolumeCopyFiles = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeCopyFiles, client)
        self.VolumeDelete = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeDelete, client)
        self.VolumeGetFile = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeGetFile, client)
        self.VolumeGetOrCreate = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeGetOrCreate, client)
        self.VolumeHeartbeat = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeHeartbeat, client)
        self.VolumeList = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeList, client)
        self.VolumeListFiles = modal.client.UnaryStreamWrapper(grpclib_stub.VolumeListFiles, client)
        self.VolumePutFiles = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumePutFiles, client)
        self.VolumeReload = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeReload, client)
        self.VolumeRemoveFile = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeRemoveFile, client)
        self.VolumeRename = modal.client.UnaryUnaryWrapper(grpclib_stub.VolumeRename, client)
        self.WorkspaceNameLookup = modal.client.UnaryUnaryWrapper(grpclib_stub.WorkspaceNameLookup, client)
